# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import threading
import functools
import numba.cuda.core.event as ev
from numba.cuda import HAS_NUMBA

if HAS_NUMBA:
    from numba.core.compiler_lock import (
        global_compiler_lock as _numba_compiler_lock,
    )
else:
    _numba_compiler_lock = None


# Lock for the preventing multiple compiler execution
class _CompilerLock:
    def __init__(self):
        self._lock = threading.RLock()

    def acquire(self):
        ev.start_event("numba-cuda:compiler_lock")
        self._lock.acquire()

    def release(self):
        self._lock.release()
        ev.end_event("numba-cuda:compiler_lock")

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_val, exc_type, traceback):
        self.release()

    def __call__(self, func):
        @functools.wraps(func)
        def _acquire_compile_lock(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return _acquire_compile_lock


_numba_cuda_compiler_lock = _CompilerLock()


# Wrapper that coordinates both numba and numba-cuda compiler locks
class _DualCompilerLock:
    """Wrapper that coordinates both the numba-cuda and upstream numba compiler locks."""

    def __init__(self, cuda_lock, numba_lock):
        self._cuda_lock = cuda_lock
        self._numba_lock = numba_lock

    def acquire(self):
        self._numba_lock.acquire()
        self._cuda_lock.acquire()

    def release(self):
        self._cuda_lock.release()
        self._numba_lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_val, exc_type, traceback):
        self.release()

    def __call__(self, func):
        @functools.wraps(func)
        def _acquire_compile_lock(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return _acquire_compile_lock


# Create the global compiler lock, wrapping both locks if numba is available
if HAS_NUMBA:
    global_compiler_lock = _DualCompilerLock(
        _numba_cuda_compiler_lock, _numba_compiler_lock
    )
else:
    global_compiler_lock = _numba_cuda_compiler_lock



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import weakref
import importlib


class Environment:
    """Stores globals and constant pyobjects for runtime.

    It is often needed to convert b/w nopython objects and pyobjects.
    """

    __slots__ = ("env_name", "__weakref__")
    # A weak-value dictionary to store live environment with env_name as the
    # key.
    _memo = weakref.WeakValueDictionary()

    @classmethod
    def from_fndesc(cls, fndesc):
        try:
            # Avoid creating new Env
            return cls._memo[fndesc.env_name]
        except KeyError:
            inst = cls()
            inst.env_name = fndesc.env_name
            cls._memo[fndesc.env_name] = inst
            return inst

    def can_cache(self):
        is_dyn = "__name__" not in self.globals
        return not is_dyn

    def __reduce__(self):
        return _rebuild_env, (
            self.globals.get("__name__"),
            self.consts,
            self.env_name,
        )

    def __del__(self):
        return

    def __repr__(self):
        return f"<Environment {self.env_name!r} >"


def _rebuild_env(modname, consts, env_name):
    env = lookup_environment(env_name)
    if env is not None:
        return env

    mod = importlib.import_module(modname)
    env = Environment(mod.__dict__)
    env.consts[:] = consts
    env.env_name = env_name
    # Cache loaded object
    Environment._memo[env_name] = env
    return env


def lookup_environment(env_name):
    """Returns the Environment object for the given name;
    or None if not found
    """
    return Environment._memo.get(env_name)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Exception handling intrinsics.
"""

from numba.cuda import types
from numba.cuda.core import errors
from numba.cuda import cgutils
from numba.cuda.extending import intrinsic


@intrinsic
def exception_check(typingctx):
    """An intrinsic to check if an exception is raised"""

    def codegen(context, builder, signature, args):
        nrt = context.nrt
        return nrt.eh_check(builder)

    restype = types.boolean
    return restype(), codegen


@intrinsic
def mark_try_block(typingctx):
    """An intrinsic to mark the start of a *try* block."""

    def codegen(context, builder, signature, args):
        nrt = context.nrt
        nrt.eh_try(builder)
        return context.get_dummy_value()

    restype = types.none
    return restype(), codegen


@intrinsic
def end_try_block(typingctx):
    """An intrinsic to mark the end of a *try* block."""

    def codegen(context, builder, signature, args):
        nrt = context.nrt
        nrt.eh_end_try(builder)
        return context.get_dummy_value()

    restype = types.none
    return restype(), codegen


@intrinsic
def exception_match(typingctx, exc_value, exc_class):
    """Basically do ``isinstance(exc_value, exc_class)`` for exception objects.
    Used in ``except Exception:`` syntax.
    """
    # Check for our limitation
    if exc_class.exc_class is not Exception:
        msg = "Exception matching is limited to {}"
        raise errors.UnsupportedError(msg.format(Exception))

    def codegen(context, builder, signature, args):
        # Intentionally always True.
        return cgutils.true_bit

    restype = types.boolean
    return restype(exc_value, exc_class), codegen



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.core.tracing import event

from numba.cuda.core import errors
from numba.cuda.core.errors import CompilerError

from numba.cuda.core import callconv, config, bytecode
from numba.cuda.core.untyped_passes import ExtractByteCode, FixupArgs
from numba.cuda.core.targetconfig import ConfigStack

from numba.cuda.core.callconv import CUDACallConv


class _CompileStatus:
    """
    Describes the state of compilation. Used like a C record.
    """

    __slots__ = ["fail_reason", "can_fallback"]

    def __init__(self, can_fallback):
        self.fail_reason = None
        self.can_fallback = can_fallback

    def __repr__(self):
        return ", ".join(
            "{k}={v}".format(k=k, v=getattr(self, k)) for k in self.__slots__
        )


class StateDict(dict):
    """
    A dictionary that has an overloaded getattr and setattr to permit getting
    and setting key/values through the use of attributes.
    """

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self[attr] = value


class _EarlyPipelineCompletion(Exception):
    """
    Raised to indicate that a pipeline has completed early
    """

    def __init__(self, result):
        self.result = result


def _make_subtarget(targetctx, flags):
    """
    Make a new target context from the given target context and flags.
    """
    subtargetoptions = {}
    if flags.debuginfo:
        subtargetoptions["enable_debuginfo"] = True
    if flags.boundscheck:
        subtargetoptions["enable_boundscheck"] = True
    if flags.nrt:
        subtargetoptions["enable_nrt"] = True
    if flags.fastmath:
        subtargetoptions["fastmath"] = flags.fastmath

    # Only the CUDA Calling convention can raise exceptions, so we assume here
    # that it is suitable for creating the error model.
    call_conv = CUDACallConv(targetctx)
    error_model = callconv.create_error_model(flags.error_model, call_conv)
    subtargetoptions["error_model"] = error_model

    return targetctx.subtarget(**subtargetoptions)


class CompilerBase:
    """
    Stores and manages states for the compiler
    """

    def __init__(
        self,
        typingctx,
        targetctx,
        library,
        args,
        return_type,
        flags,
        locals,
        call_conv=None,
        abi_info=None,
    ):
        # Make sure the environment is reloaded
        config.reload_config()
        typingctx.refresh()
        targetctx.refresh()

        self.state = StateDict()

        self.state.typingctx = typingctx
        self.state.targetctx = _make_subtarget(targetctx, flags)
        self.state.library = library
        self.state.args = args
        self.state.return_type = return_type
        self.state.flags = flags
        self.state.locals = locals

        # Results of various steps of the compilation pipeline
        self.state.bc = None
        self.state.func_id = None
        self.state.func_ir = None
        self.state.lifted = None
        self.state.lifted_from = None
        self.state.typemap = None
        self.state.calltypes = None
        self.state.type_annotation = None
        # holds arbitrary inter-pipeline stage meta data
        self.state.metadata = {}
        self.state.reload_init = []
        # hold this for e.g. with_lifting, null out on exit
        self.state.pipeline = self

        if call_conv is None:
            call_conv = CUDACallConv(self.state.targetctx)
        if abi_info is None:
            abi_info = {}

        self.state.call_conv = call_conv
        self.state.abi_info = abi_info

        self.state.status = _CompileStatus(
            can_fallback=self.state.flags.enable_pyobject
        )

    def compile_extra(self, func):
        self.state.func_id = bytecode.FunctionIdentity.from_function(func)
        ExtractByteCode().run_pass(self.state)

        self.state.lifted = ()
        self.state.lifted_from = None
        return self._compile_bytecode()

    def compile_ir(self, func_ir, lifted=(), lifted_from=None):
        self.state.func_id = func_ir.func_id
        self.state.lifted = lifted
        self.state.lifted_from = lifted_from
        self.state.func_ir = func_ir
        self.state.nargs = self.state.func_ir.arg_count

        FixupArgs().run_pass(self.state)
        return self._compile_ir()

    def define_pipelines(self):
        """Child classes override this to customize the pipelines in use."""
        raise NotImplementedError()

    def _compile_core(self):
        """
        Populate and run compiler pipeline
        """
        with ConfigStack().enter(self.state.flags.copy()):
            pms = self.define_pipelines()
            for pm in pms:
                pipeline_name = pm.pipeline_name
                func_name = "%s.%s" % (
                    self.state.func_id.modname,
                    self.state.func_id.func_qualname,
                )

                event("Pipeline: %s for %s" % (pipeline_name, func_name))
                self.state.metadata["pipeline_times"] = {
                    pipeline_name: pm.exec_times
                }
                is_final_pipeline = pm == pms[-1]
                res = None
                try:
                    pm.run(self.state)
                    if self.state.cr is not None:
                        break
                except _EarlyPipelineCompletion as e:
                    res = e.result
                    break
                except Exception as e:
                    if not isinstance(e, errors.NumbaError):
                        raise e
                    self.state.status.fail_reason = e
                    if is_final_pipeline:
                        raise e
            else:
                raise CompilerError("All available pipelines exhausted")

            # Pipeline is done, remove self reference to release refs to user
            # code
            self.state.pipeline = None

            # organise a return
            if res is not None:
                # Early pipeline completion
                return res
            else:
                assert self.state.cr is not None
                return self.state.cr

    def _compile_bytecode(self):
        """
        Populate and run pipeline for bytecode input
        """
        assert self.state.func_ir is None
        return self._compile_core()

    def _compile_ir(self):
        """
        Populate and run pipeline for IR input
        """
        assert self.state.func_ir is not None
        return self._compile_core()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.np.ufunc.decorators import vectorize, guvectorize  # noqa: F401



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools
from .manager import DataModelManager


def register(dmm, typecls):
    """Used as decorator to simplify datamodel registration.
    Returns the object being decorated so that chaining is possible.
    """

    def wraps(fn):
        dmm.register(typecls, fn)
        return fn

    return wraps


default_manager = DataModelManager()

register_default = functools.partial(register, default_manager)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.simulator.cudadrv import (
    devicearray,
    devices,
    driver,
    drvapi,
    error,
    nvvm,
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
A subpackage hosting Numba IR rewrite passes.
"""

from .registry import register_rewrite, rewrite_registry, Rewrite

# Register various built-in rewrite passes
from numba.cuda.core.rewrites import (
    static_getitem,
    static_raise,
    static_binop,
    ir_print,
)

__all__ = (
    "static_getitem",
    "static_raise",
    "static_binop",
    "ir_print",
    "register_rewrite",
    "rewrite_registry",
    "Rewrite",
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
This module provides helper functions to find the first line of a function
body.
"""

import ast
import inspect
import textwrap


class FindDefFirstLine(ast.NodeVisitor):
    """
    Attributes
    ----------
    first_stmt_line : int or None
        This stores the first statement line number if the definition is found.
        Or, ``None`` if the definition is not found.
    def_lineno : int or None
        This stores the 'def' line number if the definition is found.
        Or, ``None`` if the definition is not found.
    """

    def __init__(self, name, firstlineno):
        """
        Parameters
        ----------
        name : str
            The function's name (co_name).
        firstlineno : int
            The function's first line number (co_firstlineno), adjusted for
            any offset if the source is a fragment.
        """
        self._co_name = name
        self._co_firstlineno = firstlineno
        self.first_stmt_line = None
        self.def_lineno = None

    def _visit_children(self, node):
        for child in ast.iter_child_nodes(node):
            super().visit(child)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        if node.name == self._co_name:
            # Name of function matches.

            # The `def` line may match co_firstlineno.
            possible_start_lines = set([node.lineno])
            if node.decorator_list:
                # Has decorators.
                # The first decorator line may match co_firstlineno.
                first_decor = node.decorator_list[0]
                possible_start_lines.add(first_decor.lineno)
            # Does the first lineno match?
            if self._co_firstlineno in possible_start_lines:
                # Yes, we found the function.
                self.def_lineno = node.lineno
                # So, use the first statement line as the first line.
                if node.body:
                    first_stmt = node.body[0]
                    if _is_docstring(first_stmt):
                        # Skip docstring
                        first_stmt = node.body[1]
                    self.first_stmt_line = first_stmt.lineno
                    return
                else:
                    # This is probably unreachable.
                    # Function body cannot be bare. It must at least have
                    # A const string for docstring or a `pass`.
                    pass
        self._visit_children(node)


def _is_docstring(node):
    if isinstance(node, ast.Expr):
        if isinstance(node.value, ast.Constant) and isinstance(
            node.value.value, str
        ):
            return True
    return False


def get_func_body_first_lineno(pyfunc):
    """
    Look up the first line of function body.

    Uses inspect.getsourcelines() which works for both regular .py files
    (via linecache reading from disk) and Jupyter notebook cells (via
    IPython's linecache registration).

    Returns
    -------
    lineno : int; or None
        The first line number of the function body; or ``None`` if the first
        line cannot be determined.
    """
    co = pyfunc.__code__
    try:
        lines, offset = inspect.getsourcelines(pyfunc)
        source = "".join(lines)
        offset = offset - 1
    except (OSError, TypeError):
        return None

    tree = ast.parse(textwrap.dedent(source))
    finder = FindDefFirstLine(co.co_name, co.co_firstlineno - offset)
    finder.visit(tree)
    if finder.first_stmt_line:
        return finder.first_stmt_line + offset
    else:
        # No first line found.
        return None


def get_func_def_lineno(pyfunc):
    """
    Look up the line number of the function definition ('def' line).

    Uses inspect.getsourcelines() which works for both regular .py files
    (via linecache reading from disk) and Jupyter notebook cells (via
    IPython's linecache registration).

    Returns
    -------
    lineno : int; or None
        The line number of the function definition (the 'def' line); or
        ``None`` if it cannot be determined.
    """
    co = pyfunc.__code__
    try:
        lines, offset = inspect.getsourcelines(pyfunc)
        source = "".join(lines)
        offset = offset - 1
    except (OSError, TypeError):
        return None

    tree = ast.parse(textwrap.dedent(source))
    finder = FindDefFirstLine(co.co_name, co.co_firstlineno - offset)
    finder.visit(tree)
    if finder.def_lineno:
        return finder.def_lineno + offset
    else:
        # No def line found.
        return None



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import collections
import warnings
from functools import cached_property

from llvmlite import ir

from .abstract import DTypeSpec, IteratorType, MutableSequence, Number, Type
from .common import Buffer, Opaque, SimpleIteratorType
from numba.cuda.typeconv import Conversion
from .misc import UnicodeType
from .containers import Bytes
import numpy as np


class CharSeq(Type):
    """
    A fixed-length 8-bit character sequence.
    """

    mutable = True

    def __init__(self, count):
        self.count = count
        name = "[char x %d]" % count
        super().__init__(name)

    @property
    def key(self):
        return self.count

    def can_convert_from(self, typingctx, other):
        if isinstance(other, Bytes):
            return Conversion.safe


class UnicodeCharSeq(Type):
    """
    A fixed-length unicode character sequence.
    """

    mutable = True

    def __init__(self, count):
        self.count = count
        name = "[unichr x %d]" % count
        super().__init__(name)

    @property
    def key(self):
        return self.count

    def can_convert_to(self, typingctx, other):
        if isinstance(other, UnicodeCharSeq):
            return Conversion.safe

    def can_convert_from(self, typingctx, other):
        if isinstance(other, UnicodeType):
            # Assuming that unicode_type itemsize is not greater than
            # numpy.dtype('U1').itemsize that UnicodeCharSeq is based
            # on.
            return Conversion.safe

    def __repr__(self):
        return f"UnicodeCharSeq({self.count})"


_RecordField = collections.namedtuple(
    "_RecordField",
    "type,offset,alignment,title",
)


class Record(Type):
    """
    A Record datatype can be mapped to a NumPy structured dtype.
    A record is very flexible since it is laid out as a list of bytes.
    Fields can be mapped to arbitrary points inside it, even if they overlap.

    *fields* is a list of `(name:str, data:dict)`.
        Where `data` is `{ type: Type, offset: int }`
    *size* is an int; the record size
    *aligned* is a boolean; whether the record is ABI aligned.
    """

    mutable = True

    @classmethod
    def make_c_struct(cls, name_types):
        """Construct a Record type from a list of (name:str, type:Types).
        The layout of the structure will follow C.

        Note: only scalar types are supported currently.
        """
        from numba.cuda.core.registry import cpu_target

        ctx = cpu_target.target_context
        offset = 0
        fields = []
        lltypes = []
        for k, ty in name_types:
            if not isinstance(ty, (Number, NestedArray)):
                msg = "Only Number and NestedArray types are supported, found: {}. "
                raise TypeError(msg.format(ty))
            if isinstance(ty, NestedArray):
                datatype = ctx.data_model_manager[ty].as_storage_type()
            else:
                datatype = ctx.get_data_type(ty)
            lltypes.append(datatype)
            size = ctx.get_abi_sizeof(datatype)
            align = ctx.get_abi_alignment(datatype)
            # align
            misaligned = offset % align
            if misaligned:
                offset += align - misaligned
            fields.append(
                (
                    k,
                    {
                        "type": ty,
                        "offset": offset,
                        "alignment": align,
                    },
                )
            )
            offset += size
        # Adjust sizeof structure
        abi_size = ctx.get_abi_sizeof(ir.LiteralStructType(lltypes))
        return Record(fields, size=abi_size, aligned=True)

    def __init__(self, fields, size, aligned):
        fields = self._normalize_fields(fields)
        self.fields = dict(fields)
        self.size = size
        self.aligned = aligned

        # Create description
        descbuf = []
        fmt = "{}[type={};offset={}{}]"
        for k, infos in fields:
            extra = ""
            if infos.alignment is not None:
                extra += ";alignment={}".format(infos.alignment)
            elif infos.title is not None:
                extra += ";title={}".format(infos.title)
            descbuf.append(fmt.format(k, infos.type, infos.offset, extra))

        desc = ",".join(descbuf)
        name = "Record({};{};{})".format(desc, self.size, self.aligned)
        super().__init__(name)

        self.bitwidth = self.dtype.itemsize * 8

    @classmethod
    def _normalize_fields(cls, fields):
        """
        fields:
            [name: str,
             value: {
                 type: Type,
                 offset: int,
                 [ alignment: int ],
                 [ title : str],
             }]
        """
        res = []
        for name, infos in sorted(fields, key=lambda x: (x[1]["offset"], x[0])):
            fd = _RecordField(
                type=infos["type"],
                offset=infos["offset"],
                alignment=infos.get("alignment"),
                title=infos.get("title"),
            )
            res.append((name, fd))
        return res

    @property
    def key(self):
        # Numpy dtype equality doesn't always succeed, use the name instead
        # (https://github.com/numpy/numpy/issues/5715)
        return self.name

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)

    def __len__(self):
        """Returns the number of fields"""
        return len(self.fields)

    def offset(self, key):
        """Get the byte offset of a field from the start of the structure."""
        return self.fields[key].offset

    def typeof(self, key):
        """Get the type of a field."""
        return self.fields[key].type

    def alignof(self, key):
        """Get the specified alignment of the field.

        Since field alignment is optional, this may return None.
        """
        return self.fields[key].alignment

    def has_titles(self):
        """Returns True the record uses titles."""
        return any(fd.title is not None for fd in self.fields.values())

    def is_title(self, key):
        """Returns True if the field named *key* is a title."""
        return self.fields[key].title == key

    @property
    def members(self):
        """An ordered list of (name, type) for the fields."""
        ordered = sorted(self.fields.items(), key=lambda x: x[1].offset)
        return [(k, v.type) for k, v in ordered]

    @property
    def dtype(self):
        from numba.cuda.np.numpy_support import as_struct_dtype

        return as_struct_dtype(self)

    def can_convert_to(self, typingctx, other):
        """
        Convert this Record to the *other*.

        This method only implements width subtyping for records.
        """
        from numba.cuda.core.errors import NumbaExperimentalFeatureWarning

        if isinstance(other, Record):
            if len(other.fields) > len(self.fields):
                return
            for other_fd, self_fd in zip(
                other.fields.items(), self.fields.items()
            ):
                if not other_fd == self_fd:
                    return
            warnings.warn(
                f"{self} has been considered a subtype of {other} "
                f" This is an experimental feature.",
                category=NumbaExperimentalFeatureWarning,
            )
            return Conversion.safe

    def __repr__(self):
        fields = [
            f"('{f_name}', "
            + f"{{'type': {repr(f_info.type)}, "
            + f"'offset': {f_info.offset}, "
            + f"'alignment': {f_info.alignment}, "
            + f"'title': {f_info.title}, "
            + "}"
            + ")"
            for f_name, f_info in self.fields.items()
        ]
        fields = "[" + ", ".join(fields) + "]"
        return f"Record({fields}, {self.size}, {self.aligned})"


class DType(DTypeSpec, Opaque):
    """
    Type class associated with the `np.dtype`.

    i.e. :code:`assert type(np.dtype('int32')) == np.dtype`

    np.dtype('int32')
    """

    def __init__(self, dtype):
        assert isinstance(dtype, Type)
        self._dtype = dtype
        name = "dtype(%s)" % (dtype,)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    @property
    def dtype(self):
        return self._dtype

    def __getitem__(self, arg):
        res = super().__getitem__(arg)
        return res.copy(dtype=self.dtype)


class NumpyFlatType(SimpleIteratorType, MutableSequence):
    """
    Type class for `ndarray.flat()` objects.
    """

    def __init__(self, arrty):
        self.array_type = arrty
        yield_type = arrty.dtype
        self.dtype = yield_type
        name = "array.flat({arrayty})".format(arrayty=arrty)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.array_type


class NumpyNdEnumerateType(SimpleIteratorType):
    """
    Type class for `np.ndenumerate()` objects.
    """

    def __init__(self, arrty):
        from . import Tuple, UniTuple, intp

        self.array_type = arrty
        yield_type = Tuple((UniTuple(intp, arrty.ndim), arrty.dtype))
        name = "ndenumerate({arrayty})".format(arrayty=arrty)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.array_type


class NumpyNdIterType(IteratorType):
    """
    Type class for `np.nditer()` objects.

    The layout denotes in which order the logical shape is iterated on.
    "C" means logical order (corresponding to in-memory order in C arrays),
    "F" means reverse logical order (corresponding to in-memory order in
    F arrays).
    """

    def __init__(self, arrays):
        # Note inputs arrays can also be scalars, in which case they are
        # broadcast.
        self.arrays = tuple(arrays)
        self.layout = self._compute_layout(self.arrays)
        self.dtypes = tuple(getattr(a, "dtype", a) for a in self.arrays)
        self.ndim = max(getattr(a, "ndim", 0) for a in self.arrays)
        name = "nditer(ndim={ndim}, layout={layout}, inputs={arrays})".format(
            ndim=self.ndim, layout=self.layout, arrays=self.arrays
        )
        super().__init__(name)

    @classmethod
    def _compute_layout(cls, arrays):
        c = collections.Counter()
        for a in arrays:
            if not isinstance(a, Array):
                continue
            if a.layout in "CF" and a.ndim == 1:
                c["C"] += 1
                c["F"] += 1
            elif a.ndim >= 1:
                c[a.layout] += 1
        return "F" if c["F"] > c["C"] else "C"

    @property
    def key(self):
        return self.arrays

    @property
    def views(self):
        """
        The views yielded by the iterator.
        """
        return [Array(dtype, 0, "C") for dtype in self.dtypes]

    @property
    def yield_type(self):
        from . import BaseTuple

        views = self.views
        if len(views) > 1:
            return BaseTuple.from_types(views)
        else:
            return views[0]

    @cached_property
    def indexers(self):
        """
        A list of (kind, start_dim, end_dim, indices) where:
        - `kind` is either "flat", "indexed", "0d" or "scalar"
        - `start_dim` and `end_dim` are the dimension numbers at which
          this indexing takes place
        - `indices` is the indices of the indexed arrays in self.arrays
        """
        d = collections.OrderedDict()
        layout = self.layout
        ndim = self.ndim
        assert layout in "CF"
        for i, a in enumerate(self.arrays):
            if not isinstance(a, Array):
                indexer = ("scalar", 0, 0)
            elif a.ndim == 0:
                indexer = ("0d", 0, 0)
            else:
                if a.layout == layout or (a.ndim == 1 and a.layout in "CF"):
                    kind = "flat"
                else:
                    kind = "indexed"
                if layout == "C":
                    # If iterating in C order, broadcasting is done on the outer indices
                    indexer = (kind, ndim - a.ndim, ndim)
                else:
                    indexer = (kind, 0, a.ndim)
            d.setdefault(indexer, []).append(i)
        return list(k + (v,) for k, v in d.items())

    @cached_property
    def need_shaped_indexing(self):
        """
        Whether iterating on this iterator requires keeping track of
        individual indices inside the shape.  If False, only a single index
        over the equivalent flat shape is required, which can make the
        iterator more efficient.
        """
        for kind, start_dim, end_dim, _ in self.indexers:
            if kind in ("0d", "scalar"):
                pass
            elif kind == "flat":
                if (start_dim, end_dim) != (0, self.ndim):
                    # Broadcast flat iteration needs shaped indexing
                    # to know when to restart iteration.
                    return True
            else:
                return True
        return False


class NumpyNdIndexType(SimpleIteratorType):
    """
    Type class for `np.ndindex()` objects.
    """

    def __init__(self, ndim):
        from . import UniTuple, intp

        self.ndim = ndim
        yield_type = UniTuple(intp, self.ndim)
        name = "ndindex(ndim={ndim})".format(ndim=ndim)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.ndim


class Array(Buffer):
    """
    Type class for Numpy arrays.
    """

    def __init__(
        self, dtype, ndim, layout, readonly=False, name=None, aligned=True
    ):
        if readonly:
            self.mutable = False
        if not aligned or (isinstance(dtype, Record) and not dtype.aligned):
            self.aligned = False
        if isinstance(dtype, NestedArray):
            ndim += dtype.ndim
            dtype = dtype.dtype
        if name is None:
            type_name = "array"
            if not self.mutable:
                type_name = "readonly " + type_name
            if not self.aligned:
                type_name = "unaligned " + type_name
            name = "%s(%s, %sd, %s)" % (type_name, dtype, ndim, layout)
        super().__init__(dtype, ndim, layout, name=name)

    @property
    def mangling_args(self):
        args = [
            self.dtype,
            self.ndim,
            self.layout,
            "mutable" if self.mutable else "readonly",
            "aligned" if self.aligned else "unaligned",
        ]
        return self.__class__.__name__, args

    def copy(self, dtype=None, ndim=None, layout=None, readonly=None):
        if dtype is None:
            dtype = self.dtype
        if ndim is None:
            ndim = self.ndim
        if layout is None:
            layout = self.layout
        if readonly is None:
            readonly = not self.mutable
        return Array(
            dtype=dtype,
            ndim=ndim,
            layout=layout,
            readonly=readonly,
            aligned=self.aligned,
        )

    @property
    def key(self):
        return self.dtype, self.ndim, self.layout, self.mutable, self.aligned

    def unify(self, typingctx, other):
        """
        Unify this with the *other* Array.
        """
        # If other is array and the ndim matches
        if isinstance(other, Array) and other.ndim == self.ndim:
            # If dtype matches or other.dtype is undefined (inferred)
            if other.dtype == self.dtype or not other.dtype.is_precise():
                if self.layout == other.layout:
                    layout = self.layout
                else:
                    layout = "A"
                readonly = not (self.mutable and other.mutable)
                aligned = self.aligned and other.aligned
                return Array(
                    dtype=self.dtype,
                    ndim=self.ndim,
                    layout=layout,
                    readonly=readonly,
                    aligned=aligned,
                )

    def can_convert_to(self, typingctx, other):
        """
        Convert this Array to the *other*.
        """
        if (
            isinstance(other, Array)
            and other.ndim == self.ndim
            and other.dtype == self.dtype
        ):
            if (
                other.layout in ("A", self.layout)
                and (self.mutable or not other.mutable)
                and (self.aligned or not other.aligned)
            ):
                return Conversion.safe

    def is_precise(self):
        return self.dtype.is_precise()

    @property
    def box_type(self):
        """Returns the Python type to box to."""
        return np.ndarray

    def __repr__(self):
        return (
            f"Array({repr(self.dtype)}, {self.ndim}, '{self.layout}', "
            f"{not self.mutable}, aligned={self.aligned})"
        )


class ArrayCTypes(Type):
    """
    This is the type for `np.ndarray.ctypes`.
    """

    def __init__(self, arytype):
        # This depends on the ndim for the shape and strides attributes,
        # even though they are not implemented, yet.
        self.dtype = arytype.dtype
        self.ndim = arytype.ndim
        name = "ArrayCTypes(dtype={0}, ndim={1})".format(self.dtype, self.ndim)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype, self.ndim

    def can_convert_to(self, typingctx, other):
        """
        Convert this type to the corresponding pointer type.
        This allows passing a array.ctypes object to a C function taking
        a raw pointer.

        Note that in pure Python, the array.ctypes object can only be
        passed to a ctypes function accepting a c_void_p, not a typed
        pointer.
        """
        from . import CPointer, voidptr

        # XXX what about readonly
        if isinstance(other, CPointer) and other.dtype == self.dtype:
            return Conversion.safe
        elif other == voidptr:
            return Conversion.safe


class ArrayFlags(Type):
    """
    This is the type for `np.ndarray.flags`.
    """

    def __init__(self, arytype):
        self.array_type = arytype
        name = "ArrayFlags({0})".format(self.array_type)
        super().__init__(name)

    @property
    def key(self):
        return self.array_type


class NestedArray(Array):
    """
    A NestedArray is an array nested within a structured type (which are "void"
    type in NumPy parlance). Unlike an Array, the shape, and not just the number
    of dimensions is part of the type of a NestedArray.
    """

    def __init__(self, dtype, shape):
        if isinstance(dtype, NestedArray):
            tmp = Array(dtype.dtype, dtype.ndim, "C")
            shape += dtype.shape
            dtype = tmp.dtype
        assert dtype.bitwidth % 8 == 0, (
            "Dtype bitwidth must be a multiple of bytes"
        )
        self._shape = shape
        name = "nestedarray(%s, %s)" % (dtype, shape)
        ndim = len(shape)
        super().__init__(dtype, ndim, "C", name=name)

    @property
    def shape(self):
        return self._shape

    @property
    def nitems(self):
        l = 1
        for s in self.shape:
            l = l * s
        return l

    @property
    def size(self):
        return self.dtype.bitwidth // 8

    @property
    def strides(self):
        stride = self.size
        strides = []
        for i in reversed(self._shape):
            strides.append(stride)
            stride *= i
        return tuple(reversed(strides))

    @property
    def key(self):
        return self.dtype, self.shape

    def __repr__(self):
        return f"NestedArray({repr(self.dtype)}, {self.shape})"


class NumPyRandomBitGeneratorType(Type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "NumPyRandomBitGeneratorType"


class NumPyRandomGeneratorType(Type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "NumPyRandomGeneratorType"


class PolynomialType(Type):
    def __init__(self, coef, domain=None, window=None, n_args=1):
        super().__init__(
            name=f"PolynomialType({coef}, {domain}, {domain}, {n_args})"
        )
        self.coef = coef
        self.domain = domain
        self.window = window
        # We use n_args to keep track of the number of arguments in the
        # constructor, since the types of domain and window arguments depend on
        # that and we need that information when boxing
        self.n_args = n_args



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Former CUDA Runtime wrapper.

The toolkit version can now be obtained from NVRTC, so we don't use a binding
to the runtime anymore. This file is provided to maintain the existing API.
"""

from numba.cuda.cudadrv.nvrtc import _get_nvrtc_version

SUPPORTED_TOOLKIT_MAJOR_VERSIONS = (12, 13)


class Runtime:
    def get_version(self):
        return _get_nvrtc_version()

    def is_supported_version(self):
        """
        Returns True if the CUDA Runtime is a supported version.
        """
        return self.get_version()[0] in SUPPORTED_TOOLKIT_MAJOR_VERSIONS


runtime = Runtime()


def get_version():
    """
    Return the runtime version as a tuple of (major, minor)
    """
    return runtime.get_version()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
This module implements code highlighting of numba-cuda function annotations.
"""

from warnings import warn

warn(
    "The pretty_annotate functionality is experimental and might change API",
    FutureWarning,
)


def hllines(code, style):
    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.formatters import HtmlFormatter
    except ImportError:
        raise ImportError("please install the 'pygments' package")
    pylex = PythonLexer()
    "Given a code string, return a list of html-highlighted lines"
    hf = HtmlFormatter(noclasses=True, style=style, nowrap=True)
    res = highlight(code, pylex, hf)
    return res.splitlines()


def htlines(code, style):
    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer

        # TerminalFormatter does not support themes, Terminal256 should,
        # but seem to not work.
        from pygments.formatters import TerminalFormatter
    except ImportError:
        raise ImportError("please install the 'pygments' package")
    pylex = PythonLexer()
    "Given a code string, return a list of ANSI-highlighted lines"
    hf = TerminalFormatter(style=style)
    res = highlight(code, pylex, hf)
    return res.splitlines()


def get_ansi_template():
    try:
        from jinja2 import Template
    except ImportError:
        raise ImportError("please install the 'jinja2' package")
    return Template("""
    {%- for func_key in func_data.keys() -%}
        Function name: \x1b[34m{{func_data[func_key]['funcname']}}\x1b[39;49;00m
        {%- if func_data[func_key]['filename'] -%}
        {{'\n'}}In file: \x1b[34m{{func_data[func_key]['filename'] -}}\x1b[39;49;00m
        {%- endif -%}
        {{'\n'}}With signature: \x1b[34m{{func_key[1]}}\x1b[39;49;00m
        {{- "\n" -}}
        {%- for num, line, hl, hc in func_data[func_key]['pygments_lines'] -%}
                {{-'\n'}}{{ num}}: {{hc-}}
                {%- if func_data[func_key]['ir_lines'][num] -%}
                    {%- for ir_line, ir_line_type in func_data[func_key]['ir_lines'][num] %}
                        {{-'\n'}}--{{- ' '*func_data[func_key]['python_indent'][num]}}
                        {{- ' '*(func_data[func_key]['ir_indent'][num][loop.index0]+4)
                        }}{{ir_line }}\x1b[41m{{ir_line_type-}}\x1b[39;49;00m
                    {%- endfor -%}
                {%- endif -%}
            {%- endfor -%}
    {%- endfor -%}
    """)


def get_html_template():
    try:
        from jinja2 import Template
    except ImportError:
        raise ImportError("please install the 'jinja2' package")
    return Template("""
    <html>
    <head>
        <style>

            .annotation_table {
                color: #000000;
                font-family: monospace;
                margin: 5px;
                width: 100%;
            }

            /* override JupyterLab style */
            .annotation_table td {
                text-align: left;
                background-color: transparent;
                padding: 1px;
            }

            .annotation_table tbody tr:nth-child(even) {
                background: white;
            }

            .annotation_table code
            {
                background-color: transparent;
                white-space: normal;
            }

            /* End override JupyterLab style */

            tr:hover {
                background-color: rgba(92, 200, 249, 0.25);
            }

            td.object_tag summary ,
            td.lifted_tag summary{
                font-weight: bold;
                display: list-item;
            }

            span.lifted_tag {
                color: #00cc33;
            }

            span.object_tag {
                color: #cc3300;
            }


            td.lifted_tag {
                background-color: #cdf7d8;
            }

            td.object_tag {
                background-color: #fef5c8;
            }

            code.ir_code {
                color: grey;
                font-style: italic;
            }

            .metadata {
                border-bottom: medium solid black;
                display: inline-block;
                padding: 5px;
                width: 100%;
            }

            .annotations {
                padding: 5px;
            }

            .hidden {
                display: none;
            }

            .buttons {
                padding: 10px;
                cursor: pointer;
            }
        </style>
    </head>

    <body>
        {% for func_key in func_data.keys() %}
            <div class="metadata">
            Function name: {{func_data[func_key]['funcname']}}<br />
            {% if func_data[func_key]['filename'] %}
                in file: {{func_data[func_key]['filename']|escape}}<br />
            {% endif %}
            with signature: {{func_key[1]|e}}
            </div>
            <div class="annotations">
            <table class="annotation_table tex2jax_ignore">
                {%- for num, line, hl, hc in func_data[func_key]['pygments_lines'] -%}
                    {%- if func_data[func_key]['ir_lines'][num] %}
                        <tr><td style="text-align:left;" class="{{func_data[func_key]['python_tags'][num]}}">
                            <details>
                                <summary>
                                    <code>
                                    {{num}}:
                                    {{'&nbsp;'*func_data[func_key]['python_indent'][num]}}{{hl}}
                                    </code>
                                </summary>
                                <table class="annotation_table">
                                    <tbody>
                                        {%- for ir_line, ir_line_type in func_data[func_key]['ir_lines'][num] %}
                                            <tr class="ir_code">
                                                <td style="text-align: left;"><code>
                                                &nbsp;
                                                {{- '&nbsp;'*func_data[func_key]['python_indent'][num]}}
                                                {{ '&nbsp;'*func_data[func_key]['ir_indent'][num][loop.index0]}}{{ir_line|e -}}
                                                <span class="object_tag">{{ir_line_type}}</span>
                                                </code>
                                                </td>
                                            </tr>
                                        {%- endfor -%}
                                    </tbody>
                                </table>
                                </details>
                        </td></tr>
                    {% else -%}
                        <tr><td style="text-align:left; padding-left: 22px;" class="{{func_data[func_key]['python_tags'][num]}}">
                            <code>
                                {{num}}:
                                {{'&nbsp;'*func_data[func_key]['python_indent'][num]}}{{hl}}
                            </code>
                        </td></tr>
                    {%- endif -%}
                {%- endfor -%}
            </table>
            </div>
        {% endfor %}
    </body>
    </html>
    """)


def reform_code(annotation):
    """
    Extract the code from the Numba-cuda annotation datastructure.

    Pygments can only highlight full multi-line strings, the Numba-cuda
    annotation is list of single lines, with indentation removed.
    """
    ident_dict = annotation["python_indent"]
    s = ""
    for n, l in annotation["python_lines"]:
        s = s + " " * ident_dict[n] + l + "\n"
    return s


class Annotate:
    """
    Construct syntax highlighted annotation for a given jitted function:

    Example:

    >>> from numba import cuda
    >>> import numpy as np
    >>> from numba.cuda.core.annotations.pretty_annotate import Annotate
    >>> @cuda.jit
    ... def test(a):
    ...     tid = cuda.grid(1)
    ...     size = len(a)
    ...     if tid < size:
    ...         a[tid] = 1
    >>> test[(4), (16)](np.ones(100))
    >>> Annotate(test)

    The last line will return an HTML and/or ANSI representation that will be
    displayed accordingly in Jupyter/IPython.

    Function annotations persist across compilation for newly encountered
    type signatures and as a result annotations are shown for all signatures
    by default.

    Annotations for a specific signature can be shown by using the
    ``signature`` parameter. For the above jitted function:

    >>> test.signatures
    [(Array(float64, 1, 'C', False, aligned=True),)]
    >>> Annotate(f, signature=f.signatures[0])
    # annotation for Array(float64, 1, 'C', False, aligned=True)
    """

    def __init__(self, function, signature=None, **kwargs):
        style = kwargs.get("style", "default")
        if not function.signatures:
            raise ValueError(
                "function need to be jitted for at least one signature"
            )
        ann = function.get_annotation_info(signature=signature)
        self.ann = ann

        for v in ann.values():
            res = hllines(reform_code(v), style)
            rest = htlines(reform_code(v), style)
            v["pygments_lines"] = [
                (a, b, c, d)
                for (a, b), c, d in zip(v["python_lines"], res, rest)
            ]

    def _repr_html_(self):
        return get_html_template().render(func_data=self.ann)

    def __repr__(self):
        return get_ansi_template().render(func_data=self.ann)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause


def check_static_lib(lib):
    raise FileNotFoundError("Linking libraries not supported by cudasim")


def get_cuda_include_dir():
    raise FileNotFoundError("CUDA includes not supported by cudasim")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implements custom ufunc dispatch mechanism for non-CPU devices.
"""

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import operator
import warnings
from functools import reduce
import tokenize
import string

import numpy as np

from numba.cuda.np.ufunc.ufuncbuilder import _BaseUFuncBuilder, parse_identity
from numba.cuda import types
from numba.cuda.typing import signature
from numba.cuda.core import sigutils


def parse_signature(sig):
    """Parse generalized ufunc signature.

    NOTE: ',' (COMMA) is a delimiter; not separator.
          This means trailing comma is legal.
    """

    def stripws(s):
        return "".join(c for c in s if c not in string.whitespace)

    def tokenizer(src):
        def readline():
            yield src

        gen = readline()
        return tokenize.generate_tokens(lambda: next(gen))

    def parse(src):
        tokgen = tokenizer(src)
        while True:
            tok = next(tokgen)
            if tok[1] == "(":
                symbols = []
                while True:
                    tok = next(tokgen)
                    if tok[1] == ")":
                        break
                    elif tok[0] == tokenize.NAME:
                        symbols.append(tok[1])
                    elif tok[1] == ",":
                        continue
                    else:
                        raise ValueError('bad token in signature "%s"' % tok[1])
                yield tuple(symbols)
                tok = next(tokgen)
                if tok[1] == ",":
                    continue
                elif tokenize.ISEOF(tok[0]):
                    break
            elif tokenize.ISEOF(tok[0]):
                break
            else:
                raise ValueError('bad token in signature "%s"' % tok[1])

    ins, _, outs = stripws(sig).partition("->")
    inputs = list(parse(ins))
    outputs = list(parse(outs))

    # check that all output symbols are defined in the inputs
    isym = set()
    osym = set()
    for grp in inputs:
        isym |= set(grp)
    for grp in outputs:
        osym |= set(grp)

    diff = osym.difference(isym)
    if diff:
        raise NameError("undefined output symbols: %s" % ",".join(sorted(diff)))

    return inputs, outputs


def _broadcast_axis(a, b):
    """
    Raises
    ------
    ValueError if broadcast fails
    """
    if a == b:
        return a
    elif a == 1:
        return b
    elif b == 1:
        return a
    else:
        raise ValueError("failed to broadcast {0} and {1}".format(a, b))


def _pairwise_broadcast(shape1, shape2):
    """
    Raises
    ------
    ValueError if broadcast fails
    """
    shape1, shape2 = map(tuple, [shape1, shape2])

    while len(shape1) < len(shape2):
        shape1 = (1,) + shape1

    while len(shape1) > len(shape2):
        shape2 = (1,) + shape2

    return tuple(_broadcast_axis(a, b) for a, b in zip(shape1, shape2))


def _multi_broadcast(*shapelist):
    """
    Raises
    ------
    ValueError if broadcast fails
    """
    assert shapelist

    result = shapelist[0]
    others = shapelist[1:]
    try:
        for i, each in enumerate(others, start=1):  # noqa: B007
            result = _pairwise_broadcast(result, each)
    except ValueError:
        raise ValueError(f"failed to broadcast argument #{i:d}")
    else:
        return result


class UFuncMechanism:
    """
    Prepare ufunc arguments for vectorize.
    """

    DEFAULT_STREAM = None
    SUPPORT_DEVICE_SLICING = False

    def __init__(self, typemap, args):
        """Never used directly by user. Invoke by UFuncMechanism.call()."""
        self.typemap = typemap
        self.args = args
        nargs = len(self.args)
        self.argtypes = [None] * nargs
        self.scalarpos = []
        self.signature = None
        self.arrays = [None] * nargs

    def _fill_arrays(self):
        """
        Get all arguments in array form
        """
        for i, arg in enumerate(self.args):
            if self.is_device_array(arg):
                self.arrays[i] = self.as_device_array(arg)
            elif isinstance(arg, (int, float, complex, np.number)):
                # Is scalar
                self.scalarpos.append(i)
            else:
                self.arrays[i] = np.asarray(arg)

    def _fill_argtypes(self):
        """
        Get dtypes
        """
        for i, ary in enumerate(self.arrays):
            if ary is not None:
                dtype = ary.dtype
                if dtype is None:
                    dtype = np.asarray(ary).dtype
                self.argtypes[i] = dtype

    def _resolve_signature(self):
        """Resolve signature.
        May have ambiguous case.
        """
        matches = []
        # Resolve scalar args exact match first
        if self.scalarpos:
            # Try resolve scalar arguments
            for formaltys in self.typemap:
                match_map = []
                for i, (formal, actual) in enumerate(
                    zip(formaltys, self.argtypes)
                ):
                    if actual is None:
                        actual = np.asarray(self.args[i]).dtype

                    match_map.append(actual == formal)

                if all(match_map):
                    matches.append(formaltys)

        # No matching with exact match; try coercing the scalar arguments
        if not matches:
            matches = []
            for formaltys in self.typemap:
                all_matches = all(
                    actual is None or formal == actual
                    for formal, actual in zip(formaltys, self.argtypes)
                )
                if all_matches:
                    matches.append(formaltys)

        if not matches:
            raise TypeError(
                "No matching version.  GPU ufunc requires array "
                "arguments to have the exact types.  This behaves "
                "like regular ufunc with casting='no'."
            )

        if len(matches) > 1:
            raise TypeError(
                "Failed to resolve ufunc due to ambiguous "
                "signature. Too many untyped scalars. "
                "Use numpy dtype object to type tag."
            )

        # Try scalar arguments
        self.argtypes = matches[0]

    def _get_actual_args(self):
        """Return the actual arguments
        Casts scalar arguments to np.array.
        """
        for i in self.scalarpos:
            self.arrays[i] = np.array([self.args[i]], dtype=self.argtypes[i])

        return self.arrays

    def _broadcast(self, arys):
        """Perform numpy ufunc broadcasting"""
        shapelist = [a.shape for a in arys]
        shape = _multi_broadcast(*shapelist)

        for i, ary in enumerate(arys):
            if ary.shape == shape:
                pass

            else:
                if self.is_device_array(ary):
                    arys[i] = self.broadcast_device(ary, shape)

                else:
                    ax_differs = [
                        ax
                        for ax in range(len(shape))
                        if ax >= ary.ndim or ary.shape[ax] != shape[ax]
                    ]

                    missingdim = len(shape) - len(ary.shape)
                    strides = [0] * missingdim + list(ary.strides)

                    for ax in ax_differs:
                        strides[ax] = 0

                    strided = np.lib.stride_tricks.as_strided(
                        ary, shape=shape, strides=strides
                    )

                    arys[i] = self.force_array_layout(strided)

        return arys

    def get_arguments(self):
        """Prepare and return the arguments for the ufunc.
        Does not call to_device().
        """
        self._fill_arrays()
        self._fill_argtypes()
        self._resolve_signature()
        arys = self._get_actual_args()
        return self._broadcast(arys)

    def get_function(self):
        """Returns (result_dtype, function)"""
        return self.typemap[self.argtypes]

    def is_device_array(self, obj):
        """Is the `obj` a device array?
        Override in subclass
        """
        return False

    def as_device_array(self, obj):
        """Convert the `obj` to a device array
        Override in subclass

        Default implementation is an identity function
        """
        return obj

    def broadcast_device(self, ary, shape):
        """Handles ondevice broadcasting

        Override in subclass to add support.
        """
        raise NotImplementedError("broadcasting on device is not supported")

    def force_array_layout(self, ary):
        """Ensures array layout met device requirement.

        Override in sublcass
        """
        return ary

    @classmethod
    def call(cls, typemap, args, kws):
        """Perform the entire ufunc call mechanism."""
        # Handle keywords
        stream = kws.pop("stream", cls.DEFAULT_STREAM)
        out = kws.pop("out", None)

        if kws:
            warnings.warn("unrecognized keywords: %s" % ", ".join(kws))

        # Begin call resolution
        cr = cls(typemap, args)
        args = cr.get_arguments()
        resty, func = cr.get_function()

        outshape = args[0].shape

        # Adjust output value
        if out is not None and cr.is_device_array(out):
            out = cr.as_device_array(out)

        def attempt_ravel(a):
            if cr.SUPPORT_DEVICE_SLICING:
                raise NotImplementedError

            try:
                # Call the `.ravel()` method
                return a.ravel()
            except NotImplementedError:
                # If it is not a device array
                if not cr.is_device_array(a):
                    raise
                # For device array, retry ravel on the host by first
                # copying it back.
                else:
                    hostary = cr.to_host(a, stream).ravel()
                    return cr.to_device(hostary, stream)

        if args[0].ndim > 1:
            args = [attempt_ravel(a) for a in args]

        # Prepare argument on the device
        devarys = []
        any_device = False
        for a in args:
            if cr.is_device_array(a):
                devarys.append(a)
                any_device = True
            else:
                dev_a = cr.to_device(a, stream=stream)
                devarys.append(dev_a)

        # Launch
        shape = args[0].shape
        if out is None:
            # No output is provided
            devout = cr.allocate_device_array(shape, resty, stream=stream)

            devarys.extend([devout])
            cr.launch(func, shape[0], stream, devarys)

            if any_device:
                # If any of the arguments are on device,
                # Keep output on the device
                return devout.reshape(outshape)
            else:
                # Otherwise, transfer output back to host
                return devout.copy_to_host().reshape(outshape)

        elif cr.is_device_array(out):
            # If output is provided and it is a device array,
            # Return device array
            if out.ndim > 1:
                out = attempt_ravel(out)
            devout = out
            devarys.extend([devout])
            cr.launch(func, shape[0], stream, devarys)
            return devout.reshape(outshape)

        else:
            # If output is provided and it is a host array,
            # Return host array
            assert out.shape == shape
            assert out.dtype == resty
            devout = cr.allocate_device_array(shape, resty, stream=stream)
            devarys.extend([devout])
            cr.launch(func, shape[0], stream, devarys)
            return devout.copy_to_host(out, stream=stream).reshape(outshape)

    def to_device(self, hostary, stream):
        """Implement to device transfer
        Override in subclass
        """
        raise NotImplementedError

    def to_host(self, devary, stream):
        """Implement to host transfer
        Override in subclass
        """
        raise NotImplementedError

    def allocate_device_array(self, shape, dtype, stream):
        """Implements device allocation
        Override in subclass
        """
        raise NotImplementedError

    def launch(self, func, count, stream, args):
        """Implements device function invocation
        Override in subclass
        """
        raise NotImplementedError


def to_dtype(ty):
    if isinstance(ty, types.EnumMember):
        ty = ty.dtype
    return np.dtype(str(ty))


class DeviceVectorize(_BaseUFuncBuilder):
    def __init__(self, func, identity=None, cache=False, targetoptions=None):
        if cache:
            raise TypeError("caching is not supported")
        for opt in targetoptions or {}:
            if opt == "nopython":
                warnings.warn(
                    "nopython kwarg for cuda target is redundant",
                    RuntimeWarning,
                )
            else:
                msg = (
                    "Unrecognized options. "
                    f"cuda vectorize target does not support option: '{opt}'"
                )
                raise KeyError(msg)
        self.py_func = func
        self.identity = parse_identity(identity)
        # { arg_dtype: (return_dtype), cudakernel }
        self.kernelmap = OrderedDict()

    @property
    def pyfunc(self):
        return self.py_func

    def add(self, sig=None):
        # compile core as device function
        args, return_type = sigutils.normalize_signature(sig)
        devfnsig = signature(return_type, *args)

        funcname = self.pyfunc.__name__
        kernelsource = self._get_kernel_source(
            self._kernel_template, devfnsig, funcname
        )
        corefn, return_type = self._compile_core(devfnsig)
        glbl = self._get_globals(corefn)
        sig = signature(types.void, *([a[:] for a in args] + [return_type[:]]))
        exec(kernelsource, glbl)

        stager = glbl["__vectorized_%s" % funcname]
        kernel = self._compile_kernel(stager, sig)

        argdtypes = tuple(to_dtype(t) for t in devfnsig.args)
        resdtype = to_dtype(return_type)
        self.kernelmap[tuple(argdtypes)] = resdtype, kernel

    def build_ufunc(self):
        raise NotImplementedError

    def _get_kernel_source(self, template, sig, funcname):
        args = ["a%d" % i for i in range(len(sig.args))]
        fmts = dict(
            name=funcname,
            args=", ".join(args),
            argitems=", ".join("%s[__tid__]" % i for i in args),
        )
        return template.format(**fmts)

    def _compile_core(self, sig):
        raise NotImplementedError

    def _get_globals(self, corefn):
        raise NotImplementedError

    def _compile_kernel(self, fnobj, sig):
        raise NotImplementedError


class DeviceGUFuncVectorize(_BaseUFuncBuilder):
    def __init__(
        self,
        func,
        sig,
        identity=None,
        cache=False,
        targetoptions=None,
        writable_args=(),
    ):
        if cache:
            raise TypeError("caching is not supported")
        if writable_args:
            raise TypeError("writable_args are not supported")

        # Allow nopython flag to be set.
        if not targetoptions:
            targetoptions = {}
        if not targetoptions.pop("nopython", True):
            raise TypeError("nopython flag must be True")
        # Are there any more target options?
        if targetoptions:
            opts = ", ".join([repr(k) for k in targetoptions.keys()])
            fmt = "The following target options are not supported: {0}"
            raise TypeError(fmt.format(opts))

        self.py_func = func
        self.identity = parse_identity(identity)
        self.signature = sig
        self.inputsig, self.outputsig = parse_signature(self.signature)

        # Maps from a tuple of input_dtypes to (output_dtypes, kernel)
        self.kernelmap = OrderedDict()

    @property
    def pyfunc(self):
        return self.py_func

    def add(self, sig=None):
        indims = [len(x) for x in self.inputsig]
        outdims = [len(x) for x in self.outputsig]
        args, return_type = sigutils.normalize_signature(sig)

        # It is only valid to specify types.none as a return type, or to not
        # specify the return type (where the "Python None" is the return type)
        valid_return_type = return_type in (types.none, None)
        if not valid_return_type:
            raise TypeError(
                "guvectorized functions cannot return values: "
                f"signature {sig} specifies {return_type} return "
                "type"
            )

        funcname = self.py_func.__name__
        src = expand_gufunc_template(
            self._kernel_template, indims, outdims, funcname, args
        )

        glbls = self._get_globals(sig)

        exec(src, glbls)
        fnobj = glbls["__gufunc_{name}".format(name=funcname)]

        outertys = list(_determine_gufunc_outer_types(args, indims + outdims))
        kernel = self._compile_kernel(fnobj, sig=tuple(outertys))

        nout = len(outdims)
        dtypes = [np.dtype(str(t.dtype)) for t in outertys]
        indtypes = tuple(dtypes[:-nout])
        outdtypes = tuple(dtypes[-nout:])

        self.kernelmap[indtypes] = outdtypes, kernel

    def _compile_kernel(self, fnobj, sig):
        raise NotImplementedError

    def _get_globals(self, sig):
        raise NotImplementedError


def _determine_gufunc_outer_types(argtys, dims):
    for at, nd in zip(argtys, dims):
        if isinstance(at, types.Array):
            yield at.copy(ndim=nd + 1)
        else:
            if nd > 0:
                raise ValueError("gufunc signature mismatch: ndim>0 for scalar")
            yield types.Array(dtype=at, ndim=1, layout="A")


def expand_gufunc_template(template, indims, outdims, funcname, argtypes):
    """Expand gufunc source template"""
    argdims = indims + outdims
    argnames = ["arg{0}".format(i) for i in range(len(argdims))]
    checkedarg = "min({0})".format(
        ", ".join(["{0}.shape[0]".format(a) for a in argnames])
    )
    inputs = [
        _gen_src_for_indexing(aref, adims, atype)
        for aref, adims, atype in zip(argnames, indims, argtypes)
    ]
    outputs = [
        _gen_src_for_indexing(aref, adims, atype)
        for aref, adims, atype in zip(
            argnames[len(indims) :], outdims, argtypes[len(indims) :]
        )
    ]
    argitems = inputs + outputs
    src = template.format(
        name=funcname,
        args=", ".join(argnames),
        checkedarg=checkedarg,
        argitems=", ".join(argitems),
    )
    return src


def _gen_src_for_indexing(aref, adims, atype):
    return "{aref}[{sliced}]".format(
        aref=aref, sliced=_gen_src_index(adims, atype)
    )


def _gen_src_index(adims, atype):
    if adims > 0:
        return ",".join(["__tid__"] + [":"] * adims)
    elif isinstance(atype, types.Array) and atype.ndim - 1 == adims:
        # Special case for 0-nd in shape-signature but
        # 1d array in type signature.
        # Slice it so that the result has the same dimension.
        return "__tid__:(__tid__ + 1)"
    else:
        return "__tid__"


class GUFuncEngine:
    """Determine how to broadcast and execute a gufunc
    base on input shape and signature
    """

    @classmethod
    def from_signature(cls, signature):
        return cls(*parse_signature(signature))

    def __init__(self, inputsig, outputsig):
        # signatures
        self.sin = inputsig
        self.sout = outputsig
        # argument count
        self.nin = len(self.sin)
        self.nout = len(self.sout)

    def schedule(self, ishapes):
        if len(ishapes) != self.nin:
            raise TypeError("invalid number of input argument")

        # associate symbol values for input signature
        symbolmap = {}
        outer_shapes = []
        inner_shapes = []

        for argn, (shape, symbols) in enumerate(zip(ishapes, self.sin)):
            argn += 1  # start from 1 for human
            inner_ndim = len(symbols)
            if len(shape) < inner_ndim:
                fmt = "arg #%d: insufficient inner dimension"
                raise ValueError(fmt % (argn,))
            if inner_ndim:
                inner_shape = shape[-inner_ndim:]
                outer_shape = shape[:-inner_ndim]
            else:
                inner_shape = ()
                outer_shape = shape

            for axis, (dim, sym) in enumerate(zip(inner_shape, symbols)):
                axis += len(outer_shape)
                if sym in symbolmap:
                    if symbolmap[sym] != dim:
                        fmt = "arg #%d: shape[%d] mismatch argument"
                        raise ValueError(fmt % (argn, axis))
                symbolmap[sym] = dim

            outer_shapes.append(outer_shape)
            inner_shapes.append(inner_shape)

        # solve output shape
        oshapes = [
            tuple(map(symbolmap.__getitem__, outsig)) for outsig in self.sout
        ]

        # find the biggest outershape as looping dimension
        sizes = [reduce(operator.mul, s, 1) for s in outer_shapes]
        largest_i = np.argmax(sizes)
        loopdims = outer_shapes[largest_i]

        pinned = [False] * self.nin  # same argument for each iteration
        for i, d in enumerate(outer_shapes):
            if d != loopdims:
                if d == (1,) or d == ():
                    pinned[i] = True
                else:
                    fmt = "arg #%d: outer dimension mismatch"
                    raise ValueError(fmt % (i + 1,))

        return GUFuncSchedule(self, inner_shapes, oshapes, loopdims, pinned)


class GUFuncSchedule:
    def __init__(self, parent, ishapes, oshapes, loopdims, pinned):
        self.parent = parent
        # core shapes
        self.ishapes = ishapes
        self.oshapes = oshapes
        # looping dimension
        self.loopdims = loopdims
        self.loopn = reduce(operator.mul, loopdims, 1)
        # flags
        self.pinned = pinned

        self.output_shapes = [loopdims + s for s in oshapes]

    def __str__(self):
        import pprint

        attrs = "ishapes", "oshapes", "loopdims", "loopn", "pinned"
        values = [(k, getattr(self, k)) for k in attrs]
        return pprint.pformat(dict(values))


class GeneralizedUFunc:
    def __init__(self, kernelmap, engine):
        self.kernelmap = kernelmap
        self.engine = engine
        self.max_blocksize = 2**30

    def __call__(self, *args, **kws):
        callsteps = self._call_steps(
            self.engine.nin, self.engine.nout, args, kws
        )
        indtypes, schedule, outdtypes, kernel = self._schedule(
            callsteps.inputs, callsteps.outputs
        )
        callsteps.adjust_input_types(indtypes)

        outputs = callsteps.prepare_outputs(schedule, outdtypes)
        inputs = callsteps.prepare_inputs()
        parameters = self._broadcast(schedule, inputs, outputs)

        callsteps.launch_kernel(kernel, schedule.loopn, parameters)

        return callsteps.post_process_outputs(outputs)

    def _schedule(self, inputs, outs):
        input_shapes = [a.shape for a in inputs]
        schedule = self.engine.schedule(input_shapes)

        # find kernel
        indtypes = tuple(i.dtype for i in inputs)
        try:
            outdtypes, kernel = self.kernelmap[indtypes]
        except KeyError:
            # No exact match, then use the first compatible.
            # This does not match the numpy dispatching exactly.
            # Later, we may just jit a new version for the missing signature.
            indtypes = self._search_matching_signature(indtypes)
            # Select kernel
            outdtypes, kernel = self.kernelmap[indtypes]

        # check output
        for sched_shape, out in zip(schedule.output_shapes, outs):
            if out is not None and sched_shape != out.shape:
                raise ValueError("output shape mismatch")

        return indtypes, schedule, outdtypes, kernel

    def _search_matching_signature(self, idtypes):
        """
        Given the input types in `idtypes`, return a compatible sequence of
        types that is defined in `kernelmap`.

        Note: Ordering is guaranteed by `kernelmap` being a OrderedDict
        """
        for sig in self.kernelmap.keys():
            if all(
                np.can_cast(actual, desired)
                for actual, desired in zip(sig, idtypes)
            ):
                return sig
        else:
            raise TypeError("no matching signature")

    def _broadcast(self, schedule, params, retvals):
        assert schedule.loopn > 0, "zero looping dimension"

        odim = 1 if not schedule.loopdims else schedule.loopn
        newparams = []
        for p, cs in zip(params, schedule.ishapes):
            if not cs and p.size == 1:
                # Broadcast scalar input
                devary = self._broadcast_scalar_input(p, odim)
                newparams.append(devary)
            else:
                # Broadcast vector input
                newparams.append(self._broadcast_array(p, odim, cs))

        newretvals = []
        for retval, oshape in zip(retvals, schedule.oshapes):
            newretvals.append(retval.reshape(odim, *oshape))
        return tuple(newparams) + tuple(newretvals)

    def _broadcast_array(self, ary, newdim, innerdim):
        newshape = (newdim,) + innerdim
        # No change in shape
        if ary.shape == newshape:
            return ary

        # Creating new dimension
        elif len(ary.shape) < len(newshape):
            assert newshape[-len(ary.shape) :] == ary.shape, (
                "cannot add dim and reshape at the same time"
            )
            return self._broadcast_add_axis(ary, newshape)

        # Collapsing dimension
        else:
            return ary.reshape(*newshape)

    def _broadcast_add_axis(self, ary, newshape):
        raise NotImplementedError("cannot add new axis")

    def _broadcast_scalar_input(self, ary, shape):
        raise NotImplementedError


class GUFuncCallSteps(metaclass=ABCMeta):
    """
    Implements memory management and kernel launch operations for GUFunc calls.

    One instance of this class is instantiated for each call, and the instance
    is specific to the arguments given to the GUFunc call.

    The base class implements the overall logic; subclasses provide
    target-specific implementations of individual functions.
    """

    # The base class uses these slots; subclasses may provide additional slots.
    __slots__ = [
        "outputs",
        "inputs",
        "_copy_result_to_host",
    ]

    @abstractmethod
    def launch_kernel(self, kernel, nelem, args):
        """Implement the kernel launch"""

    @abstractmethod
    def is_device_array(self, obj):
        """
        Return True if `obj` is a device array for this target, False
        otherwise.
        """

    @abstractmethod
    def as_device_array(self, obj):
        """
        Return `obj` as a device array on this target.

        May return `obj` directly if it is already on the target.
        """

    @abstractmethod
    def to_device(self, hostary):
        """
        Copy `hostary` to the device and return the device array.
        """

    @abstractmethod
    def allocate_device_array(self, shape, dtype):
        """
        Allocate a new uninitialized device array with the given shape and
        dtype.
        """

    def __init__(self, nin, nout, args, kwargs):
        outputs = kwargs.get("out")

        # Ensure the user has passed a correct number of arguments
        if outputs is None and len(args) not in (nin, (nin + nout)):

            def pos_argn(n):
                return f"{n} positional argument{'s' * (n != 1)}"

            msg = (
                f"This gufunc accepts {pos_argn(nin)} (when providing "
                f"input only) or {pos_argn(nin + nout)} (when providing "
                f"input and output). Got {pos_argn(len(args))}."
            )
            raise TypeError(msg)

        if outputs is not None and len(args) > nin:
            raise ValueError(
                "cannot specify argument 'out' as both positional and keyword"
            )
        else:
            # If the user did not pass outputs either in the out kwarg or as
            # positional arguments, then we need to generate an initial list of
            # "placeholder" outputs using None as a sentry value
            outputs = [outputs] * nout

        # Ensure all output device arrays are Numba device arrays - for
        # example, any output passed in that supports the CUDA Array Interface
        # is converted to a Numba CUDA device array; others are left untouched.
        all_user_outputs_are_host = True
        self.outputs = []
        for output in outputs:
            if self.is_device_array(output):
                self.outputs.append(self.as_device_array(output))
                all_user_outputs_are_host = False
            else:
                self.outputs.append(output)

        all_host_arrays = not any([self.is_device_array(a) for a in args])

        # - If any of the arguments are device arrays, we leave the output on
        #   the device.
        self._copy_result_to_host = (
            all_host_arrays and all_user_outputs_are_host
        )

        # Normalize arguments - ensure they are either device- or host-side
        # arrays (as opposed to lists, tuples, etc).
        def normalize_arg(a):
            if self.is_device_array(a):
                convert = self.as_device_array
            else:
                convert = np.asarray

            return convert(a)

        normalized_args = [normalize_arg(a) for a in args]
        self.inputs = normalized_args[:nin]

        # Check if there are extra arguments for outputs.
        unused_inputs = normalized_args[nin:]
        if unused_inputs:
            self.outputs = unused_inputs

    def adjust_input_types(self, indtypes):
        """
        Attempt to cast the inputs to the required types if necessary
        and if they are not device arrays.

        Side effect: Only affects the elements of `inputs` that require
        a type cast.
        """
        for i, (ity, val) in enumerate(zip(indtypes, self.inputs)):
            if ity != val.dtype:
                if not hasattr(val, "astype"):
                    msg = (
                        "compatible signature is possible by casting but "
                        "{0} does not support .astype()"
                    ).format(type(val))
                    raise TypeError(msg)
                # Cast types
                self.inputs[i] = val.astype(ity)

    def prepare_outputs(self, schedule, outdtypes):
        """
        Returns a list of output parameters that all reside on the target
        device.

        Outputs that were passed-in to the GUFunc are used if they reside on the
        device; other outputs are allocated as necessary.
        """
        outputs = []
        for shape, dtype, output in zip(
            schedule.output_shapes, outdtypes, self.outputs
        ):
            if output is None or self._copy_result_to_host:
                output = self.allocate_device_array(shape, dtype)
            outputs.append(output)

        return outputs

    def prepare_inputs(self):
        """
        Returns a list of input parameters that all reside on the target device.
        """

        def ensure_device(parameter):
            if self.is_device_array(parameter):
                convert = self.as_device_array
            else:
                convert = self.to_device

            return convert(parameter)

        return [ensure_device(p) for p in self.inputs]

    def post_process_outputs(self, outputs):
        """
        Moves the given output(s) to the host if necessary.

        Returns a single value (e.g. an array) if there was one output, or a
        tuple of arrays if there were multiple. Although this feels a little
        jarring, it is consistent with the behavior of GUFuncs in general.
        """
        if self._copy_result_to_host:
            outputs = [
                self.to_host(output, self_output)
                for output, self_output in zip(outputs, self.outputs)
            ]
        elif self.outputs[0] is not None:
            outputs = self.outputs

        if len(outputs) == 1:
            return outputs[0]
        else:
            return tuple(outputs)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
drvapi is not implemented in the simulator, but this module exists to allow
tests to import correctly.
"""



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
from numba.cuda.utils import redirect_numba_module

sys.modules[__name__] = redirect_numba_module(
    locals(), "numba.core.types.containers", "numba.cuda.types.cuda_containers"
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
The compiler is not implemented in the simulator. This module provides a stub
to allow tests to import successfully.
"""

compile = None
compile_for_current_device = None
compile_ptx = None
compile_ptx_for_current_device = None
declare_device_function = None
compile_all = None


def run_frontend(func):
    pass


class DefaultPassBuilder:
    @staticmethod
    def define_nopython_lowering_pipeline(state, name="nopython_lowering"):
        pass

    @staticmethod
    def define_typed_pipeline(state, name="typed"):
        pass


class CompilerBase:
    def __init__(
        self,
        typingctx,
        targetctx,
        library,
        args,
        return_type,
        flags,
        locals,
        call_conv,
        abi_info,
    ):
        pass


PassManager = None



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools

from llvmlite import ir

from numba.cuda.datamodel.registry import DataModelManager, register
from numba.cuda.datamodel import PrimitiveModel
from numba.cuda.datamodel.models import StructModel
from numba.cuda.extending import core_models as models
from numba.cuda import types
from numba.cuda.types.ext_types import Dim3, GridGroup, CUDADispatcher, Bfloat16


cuda_data_manager = DataModelManager()

register_model = functools.partial(register, cuda_data_manager)


@register_model(Dim3)
class Dim3Model(StructModel):
    def __init__(self, dmm, fe_type):
        members = [("x", types.int32), ("y", types.int32), ("z", types.int32)]
        super().__init__(dmm, fe_type, members)


@register_model(GridGroup)
class GridGroupModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        be_type = ir.IntType(64)
        super().__init__(dmm, fe_type, be_type)


@register_model(types.Float)
class FloatModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        if fe_type == types.float16:
            be_type = ir.IntType(16)
        elif fe_type == types.float32:
            be_type = ir.FloatType()
        elif fe_type == types.float64:
            be_type = ir.DoubleType()
        else:
            raise NotImplementedError(fe_type)
        super().__init__(dmm, fe_type, be_type)


register_model(CUDADispatcher)(models.OpaqueModel)


@register_model(Bfloat16)
class _model___nv_bfloat16(PrimitiveModel):
    def __init__(self, dmm, fe_type):
        be_type = ir.IntType(16)
        super().__init__(dmm, fe_type, be_type)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""Launch configuration access for CUDA compilation contexts.

The current launch configuration is populated only during CUDA compilation
triggered by kernel launches. It is thread-local and cleared immediately after
compilation completes.
"""

import contextlib
from functools import wraps
from types import MethodType

from numba.cuda.cext import _dispatcher


def current_launch_config():
    """Return the current launch configuration, or None if not set."""
    return _dispatcher.get_current_launch_config()


def ensure_current_launch_config():
    """Return the current launch configuration or raise if not set."""
    config = current_launch_config()
    if config is None:
        raise RuntimeError("No launch config set for this thread")
    return config


@contextlib.contextmanager
def capture_compile_config(dispatcher):
    """Capture the launch config seen during compilation for a dispatcher.

    The returned dict has a single key, ``"config"``, which is populated when a
    compilation is triggered by a kernel launch. If the kernel is already
    compiled, the dict value may remain ``None``.
    """
    if dispatcher is None:
        raise TypeError("dispatcher is required")

    record = {"config": None}
    missing = object()
    original_attr = dispatcher.__dict__.get("_compile_for_args", missing)
    original = dispatcher._compile_for_args

    @wraps(original)
    def wrapped(self, *args, **kws):
        record["config"] = current_launch_config()
        return original(*args, **kws)

    dispatcher._compile_for_args = MethodType(wrapped, dispatcher)
    try:
        yield record
    finally:
        if original_attr is missing:
            del dispatcher._compile_for_args
        else:
            dispatcher._compile_for_args = original_attr


__all__ = [
    "current_launch_config",
    "ensure_current_launch_config",
    "capture_compile_config",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

try:
    from pygments.styles.default import DefaultStyle
except ImportError:
    msg = "Please install pygments to see highlighted dumps"
    raise ImportError(msg)

import numba.cuda.config
from pygments.styles.manni import ManniStyle
from pygments.styles.monokai import MonokaiStyle
from pygments.styles.native import NativeStyle

from pygments.token import Name

from pygments.style import Style


def by_colorscheme():
    """
    Get appropriate style for highlighting according to
    NUMBA_COLOR_SCHEME setting
    """
    styles = DefaultStyle.styles.copy()
    styles.update(
        {
            Name.Variable: "#888888",
        }
    )
    custom_default = type("CustomDefaultStyle", (Style,), {"styles": styles})

    style_map = {
        "no_color": custom_default,
        "dark_bg": MonokaiStyle,
        "light_bg": ManniStyle,
        "blue_bg": NativeStyle,
        "jupyter_nb": DefaultStyle,
    }

    return style_map[numba.cuda.config.COLOR_SCHEME]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import cmath

from numba.cuda import types
from numba.cuda.typing.templates import ConcreteTemplate, signature, Registry

registry = Registry()
infer_global = registry.register_global

# TODO: support non-complex arguments (floats and ints)


@infer_global(cmath.acos)
@infer_global(cmath.asin)
@infer_global(cmath.asinh)
@infer_global(cmath.atan)
@infer_global(cmath.atanh)
@infer_global(cmath.cos)
@infer_global(cmath.exp)
@infer_global(cmath.sin)
@infer_global(cmath.sqrt)
@infer_global(cmath.tan)
class CMath_unary(ConcreteTemplate):
    cases = [signature(tp, tp) for tp in sorted(types.complex_domain)]


@infer_global(cmath.isinf)
@infer_global(cmath.isnan)
class CMath_predicate(ConcreteTemplate):
    cases = [
        signature(types.boolean, tp) for tp in sorted(types.complex_domain)
    ]


@infer_global(cmath.isfinite)
class CMath_isfinite(CMath_predicate):
    pass


@infer_global(cmath.log)
class Cmath_log(ConcreteTemplate):
    # unary cmath.log()
    cases = [signature(tp, tp) for tp in sorted(types.complex_domain)]
    # binary cmath.log()
    cases += [signature(tp, tp, tp) for tp in sorted(types.complex_domain)]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.extending import (
    core_models,
    register_model,
    type_callable,
    unbox,
    NativeValue,
    make_attribute_wrapper,
    box,
)
from numba.cuda import types
from numba.cuda import cgutils
import warnings
from numba.cuda.core.errors import (
    NumbaExperimentalFeatureWarning,
    NumbaValueError,
)
from numpy.polynomial.polynomial import Polynomial
from contextlib import ExitStack
import numpy as np
from llvmlite import ir
from numba.cuda.core.imputils import Registry

registry = Registry("np.polynomial_core")
lower = registry.lower


@register_model(types.PolynomialType)
class PolynomialModel(core_models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("coef", fe_type.coef),
            ("domain", fe_type.domain),
            ("window", fe_type.window),
            # Introduced in NumPy 1.24, maybe leave it out for now
            # ('symbol', types.string)
        ]
        super().__init__(dmm, fe_type, members)


@type_callable(Polynomial)
def type_polynomial(context):
    def typer(coef, domain=None, window=None):
        default_domain = types.Array(types.int64, 1, "C")
        double_domain = types.Array(types.double, 1, "C")
        default_window = types.Array(types.int64, 1, "C")
        double_window = types.Array(types.double, 1, "C")
        double_coef = types.Array(types.double, 1, "C")

        warnings.warn(
            "Polynomial class is experimental",
            category=NumbaExperimentalFeatureWarning,
        )

        if isinstance(coef, types.Array) and all(
            [a is None for a in (domain, window)]
        ):
            if coef.ndim == 1:
                # If Polynomial(coef) is called, coef is cast to double dtype,
                # and domain and window are set to equal [-1, 1], i.e. have
                # integer dtype
                return types.PolynomialType(
                    double_coef, default_domain, default_window, 1
                )
            else:
                msg = "Coefficient array is not 1-d"
                raise NumbaValueError(msg)
        elif all([isinstance(a, types.Array) for a in (coef, domain, window)]):
            if coef.ndim == 1:
                if all([a.ndim == 1 for a in (domain, window)]):
                    # If Polynomial(coef, domain, window) is called, then coef,
                    # domain and window are cast to double dtype
                    return types.PolynomialType(
                        double_coef, double_domain, double_window, 3
                    )
            else:
                msg = "Coefficient array is not 1-d"
                raise NumbaValueError(msg)

    return typer


make_attribute_wrapper(types.PolynomialType, "coef", "coef")
make_attribute_wrapper(types.PolynomialType, "domain", "domain")
make_attribute_wrapper(types.PolynomialType, "window", "window")
# Introduced in NumPy 1.24, maybe leave it out for now
# make_attribute_wrapper(types.PolynomialType, 'symbol', 'symbol')


@lower(Polynomial, types.Array)
def impl_polynomial1(context, builder, sig, args):
    def to_double(arr):
        return np.asarray(arr, dtype=np.double)

    def const_impl():
        return np.asarray([-1, 1])

    typ = sig.return_type
    polynomial = cgutils.create_struct_proxy(typ)(context, builder)
    sig_coef = sig.args[0].copy(dtype=types.double)(sig.args[0])
    coef_cast = context.compile_internal(builder, to_double, sig_coef, args)
    sig_domain = sig.args[0].copy(dtype=types.intp)()
    sig_window = sig.args[0].copy(dtype=types.intp)()
    domain_cast = context.compile_internal(builder, const_impl, sig_domain, ())
    window_cast = context.compile_internal(builder, const_impl, sig_window, ())
    polynomial.coef = coef_cast
    polynomial.domain = domain_cast
    polynomial.window = window_cast

    return polynomial._getvalue()


@lower(Polynomial, types.Array, types.Array, types.Array)
def impl_polynomial3(context, builder, sig, args):
    def to_double(coef):
        return np.asarray(coef, dtype=np.double)

    typ = sig.return_type
    polynomial = cgutils.create_struct_proxy(typ)(context, builder)

    coef_sig = sig.args[0].copy(dtype=types.double)(sig.args[0])
    domain_sig = sig.args[1].copy(dtype=types.double)(sig.args[1])
    window_sig = sig.args[2].copy(dtype=types.double)(sig.args[2])
    coef_cast = context.compile_internal(
        builder, to_double, coef_sig, (args[0],)
    )
    domain_cast = context.compile_internal(
        builder, to_double, domain_sig, (args[1],)
    )
    window_cast = context.compile_internal(
        builder, to_double, window_sig, (args[2],)
    )

    domain_helper = context.make_helper(
        builder, domain_sig.return_type, value=domain_cast
    )
    window_helper = context.make_helper(
        builder, window_sig.return_type, value=window_cast
    )

    i64 = ir.IntType(64)
    two = i64(2)

    s1 = builder.extract_value(domain_helper.shape, 0)
    s2 = builder.extract_value(window_helper.shape, 0)
    pred1 = builder.icmp_signed("!=", s1, two)
    pred2 = builder.icmp_signed("!=", s2, two)

    with cgutils.if_unlikely(builder, pred1):
        context.fndesc.call_conv.return_user_exc(
            builder,
            ValueError,
            ("Domain has wrong number of elements.",),
        )

    with cgutils.if_unlikely(builder, pred2):
        context.fndesc.call_conv.return_user_exc(
            builder,
            ValueError,
            ("Window has wrong number of elements.",),
        )

    polynomial.coef = coef_cast
    polynomial.domain = domain_helper._getvalue()
    polynomial.window = window_helper._getvalue()

    return polynomial._getvalue()


@unbox(types.PolynomialType)
def unbox_polynomial(typ, obj, c):
    """
    Convert a Polynomial object to a native polynomial structure.
    """
    is_error_ptr = cgutils.alloca_once_value(c.builder, cgutils.false_bit)
    polynomial = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    with ExitStack() as stack:
        natives = []
        for name in ("coef", "domain", "window"):
            attr = c.pyapi.object_getattr_string(obj, name)
            with cgutils.early_exit_if_null(c.builder, stack, attr):
                c.builder.store(cgutils.true_bit, is_error_ptr)
            t = getattr(typ, name)
            native = c.unbox(t, attr)
            c.pyapi.decref(attr)
            with cgutils.early_exit_if(c.builder, stack, native.is_error):
                c.builder.store(cgutils.true_bit, is_error_ptr)
            natives.append(native)

        polynomial.coef = natives[0]
        polynomial.domain = natives[1]
        polynomial.window = natives[2]

    return NativeValue(
        polynomial._getvalue(), is_error=c.builder.load(is_error_ptr)
    )


@box(types.PolynomialType)
def box_polynomial(typ, val, c):
    """
    Convert a native polynomial structure to a Polynomial object.
    """
    ret_ptr = cgutils.alloca_once(c.builder, c.pyapi.pyobj)
    fail_obj = c.pyapi.get_null_object()

    with ExitStack() as stack:
        polynomial = cgutils.create_struct_proxy(typ)(
            c.context, c.builder, value=val
        )
        coef_obj = c.box(typ.coef, polynomial.coef)
        with cgutils.early_exit_if_null(c.builder, stack, coef_obj):
            c.builder.store(fail_obj, ret_ptr)

        domain_obj = c.box(typ.domain, polynomial.domain)
        with cgutils.early_exit_if_null(c.builder, stack, domain_obj):
            c.builder.store(fail_obj, ret_ptr)

        window_obj = c.box(typ.window, polynomial.window)
        with cgutils.early_exit_if_null(c.builder, stack, window_obj):
            c.builder.store(fail_obj, ret_ptr)

        class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(Polynomial))
        with cgutils.early_exit_if_null(c.builder, stack, class_obj):
            c.pyapi.decref(coef_obj)
            c.pyapi.decref(domain_obj)
            c.pyapi.decref(window_obj)
            c.builder.store(fail_obj, ret_ptr)

        if typ.n_args == 1:
            res1 = c.pyapi.call_function_objargs(class_obj, (coef_obj,))
            c.builder.store(res1, ret_ptr)
        else:
            res3 = c.pyapi.call_function_objargs(
                class_obj, (coef_obj, domain_obj, window_obj)
            )
            c.builder.store(res3, ret_ptr)

        c.pyapi.decref(coef_obj)
        c.pyapi.decref(domain_obj)
        c.pyapi.decref(window_obj)
        c.pyapi.decref(class_obj)

    return c.builder.load(ret_ptr)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from . import Dispatcher, Number, Integer, Float
from .abstract import Type

from numba.cuda.typeconv import Conversion


class Dim3(Type):
    """
    A 3-tuple (x, y, z) representing the position of a block or thread.
    """

    def __init__(self):
        super().__init__(name="Dim3")


class GridGroup(Type):
    """
    The grid of all threads in a cooperative kernel launch.
    """

    def __init__(self):
        super().__init__(name="GridGroup")


dim3 = Dim3()
grid_group = GridGroup()


class CUDADispatcher(Dispatcher):
    """The type of CUDA dispatchers"""

    # This type exists (instead of using types.Dispatcher as the type of CUDA
    # dispatchers) so that we can have an alternative lowering for them to the
    # lowering of CPU dispatchers - the CPU target lowers all dispatchers as a
    # constant address, but we need to lower to a dummy value because it's not
    # generally valid to use the address of CUDA kernels and functions.
    #
    # Notes: it may be a bug in the CPU target that it lowers all dispatchers to
    # a constant address - it should perhaps only lower dispatchers acting as
    # first-class functions to a constant address. Even if that bug is fixed, it
    # is still probably a good idea to have a separate type for CUDA
    # dispatchers, and this type might get other differentiation from the CPU
    # dispatcher type in future.


class Bfloat16(Number):
    """
    A bfloat16 type. Has 8 exponent bits and 7 significand bits.

    Conversion rules:
    Floats:
    from:
        fp32, fp64: UNSAFE
        fp16: UNSAFE (loses precision)
    to:
        fp32, fp64: PROMOTE (same exponent, more mantissa)
        fp16: UNSAFE (loses range)

    Integers:
    from:
        int8: SAFE
        other int: All UNSAFE (bf16 cannot represent all integers in range)
    to: UNSAFE (loses precision, round to zeros)

    All other conversions are not allowed.
    """

    def __init__(self):
        super().__init__(name="__nv_bfloat16")

        self.alignof_ = 2
        self.bitwidth = 16

    def can_convert_from(self, typingctx, other):
        if isinstance(other, Float):
            return Conversion.unsafe

        elif isinstance(other, Integer):
            if other.bitwidth == 8:
                return Conversion.safe
            else:
                return Conversion.unsafe

    def can_convert_to(self, typingctx, other):
        if isinstance(other, Float):
            if other.bitwidth >= 32:
                return Conversion.safe
            else:
                return Conversion.unsafe
        elif isinstance(other, Integer):
            return Conversion.unsafe

    def unify(self, typingctx, other):
        if isinstance(other, (Float, Integer)):
            return typingctx.unify_pairs(self, other)


bfloat16 = Bfloat16()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
The runtime API is unsupported in the simulator, but some stubs are
provided to allow tests to import correctly.
"""


class FakeRuntime:
    def get_version(self):
        return (-1, -1)

    def is_supported_version(self):
        return True

    @property
    def supported_versions(self):
        return ((-1, -1),)


runtime = FakeRuntime()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import importlib
import importlib.abc
import pathlib
import sys
import warnings

multiple_locations_msg = (
    "Multiple submodule search locations for {}. "
    "Cannot redirect numba.cuda to numba_cuda"
)

no_spec_msg = (
    "Couldn't get spec for {}. Cannot redirect numba.cuda to numba_cuda"
)


class NumbaCudaFinder(importlib.abc.MetaPathFinder):
    def __init__(self):
        self.initialized = None

    def ensure_initialized(self):
        if self.initialized is not None:
            return self.initialized

        numba_spec = importlib.util.find_spec("numba")

        if numba_spec is None:
            warnings.warn(no_spec_msg.format("numba"))
            self.initialized = False
            return False

        numba_cuda_spec = importlib.util.find_spec("numba_cuda")

        if numba_cuda_spec is None:
            warnings.warn(no_spec_msg.format("numba_cuda"))
            self.initialized = False
            return False

        numba_search_locations = numba_spec.submodule_search_locations
        numba_cuda_search_locations = numba_cuda_spec.submodule_search_locations

        if len(numba_search_locations) != 1:
            warnings.warn(multiple_locations_msg.format("numba"))
            self.initialized = False
            return False

        if len(numba_cuda_search_locations) != 1:
            warnings.warn(multiple_locations_msg.format("numba_cuda"))
            self.initialized = False
            return False

        self.numba_path = numba_search_locations[0]

        location = numba_cuda_search_locations[0]
        self.numba_cuda_path = str((pathlib.Path(location) / "numba"))

        self.initialized = True
        return True

    def find_spec(self, name, path, target=None):
        if "numba.cuda" in name:
            initialized = self.ensure_initialized()
            if not initialized:
                return None

            if any(self.numba_cuda_path in p for p in path):
                # Re-entrancy - return and carry on
                return None

            oot_path = [
                p.replace(self.numba_path, self.numba_cuda_path) for p in path
            ]
            for finder in sys.meta_path:
                try:
                    spec = finder.find_spec(name, oot_path, target)
                except AttributeError:
                    # Finders written to a pre-Python 3.4 spec for finders will
                    # not implement find_spec. We can skip those altogether.
                    continue
                else:
                    if spec is not None:
                        return spec


finder = NumbaCudaFinder()
sys.meta_path.insert(0, finder)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implement transformation on Numba IR
"""

from collections import namedtuple, defaultdict
import logging
import operator

from numba.cuda.core.analysis import (
    compute_cfg_from_blocks,
    find_top_level_loops,
)
from numba.cuda.core import ir
from numba.cuda.core import errors
from numba.cuda.core import ir_utils
from numba.cuda.core.analysis import compute_use_defs


_logger = logging.getLogger(__name__)


def _extract_loop_lifting_candidates(cfg, blocks):
    """
    Returns a list of loops that are candidate for loop lifting
    """

    # check well-formed-ness of the loop
    def same_exit_point(loop):
        "all exits must point to the same location"
        outedges = set()
        for k in loop.exits:
            succs = set(x for x, _ in cfg.successors(k))
            if not succs:
                # If the exit point has no successor, it contains an return
                # statement, which is not handled by the looplifting code.
                # Thus, this loop is not a candidate.
                _logger.debug("return-statement in loop.")
                return False
            outedges |= succs
        ok = len(outedges) == 1
        _logger.debug("same_exit_point=%s (%s)", ok, outedges)
        return ok

    def one_entry(loop):
        "there is one entry"
        ok = len(loop.entries) == 1
        _logger.debug("one_entry=%s", ok)
        return ok

    def cannot_yield(loop):
        "cannot have yield inside the loop"
        insiders = set(loop.body) | set(loop.entries) | set(loop.exits)
        for blk in map(blocks.__getitem__, insiders):
            for inst in blk.body:
                if isinstance(inst, ir.assign_types):
                    if isinstance(inst.value, ir.yield_types):
                        _logger.debug("has yield")
                        return False
        _logger.debug("no yield")
        return True

    _logger.info("finding looplift candidates")
    # the check for cfg.entry_point in the loop.entries is to prevent a bad
    # rewrite where a prelude for a lifted loop would get written into block -1
    # if a loop entry were in block 0
    candidates = []
    for loop in find_top_level_loops(cfg):
        _logger.debug("top-level loop: %s", loop)
        if (
            same_exit_point(loop)
            and one_entry(loop)
            and cannot_yield(loop)
            and cfg.entry_point() not in loop.entries
        ):
            candidates.append(loop)
            _logger.debug("add candidate: %s", loop)
    return candidates


def find_region_inout_vars(blocks, livemap, callfrom, returnto, body_block_ids):
    """Find input and output variables to a block region."""
    inputs = livemap[callfrom]
    outputs = livemap[returnto]

    # ensure live variables are actually used in the blocks, else remove,
    # saves having to create something valid to run through postproc
    # to achieve similar
    loopblocks = {}
    for k in body_block_ids:
        loopblocks[k] = blocks[k]

    used_vars = set()
    def_vars = set()
    defs = compute_use_defs(loopblocks)
    for vs in defs.usemap.values():
        used_vars |= vs
    for vs in defs.defmap.values():
        def_vars |= vs
    used_or_defined = used_vars | def_vars

    # note: sorted for stable ordering
    inputs = sorted(set(inputs) & used_or_defined)
    outputs = sorted(set(outputs) & used_or_defined & def_vars)
    return inputs, outputs


_loop_lift_info = namedtuple(
    "loop_lift_info", "loop,inputs,outputs,callfrom,returnto"
)


def _loop_lift_get_candidate_infos(cfg, blocks, livemap):
    """
    Returns information on looplifting candidates.
    """
    loops = _extract_loop_lifting_candidates(cfg, blocks)
    loopinfos = []
    for loop in loops:
        [callfrom] = loop.entries  # requirement checked earlier
        an_exit = next(iter(loop.exits))  # anyone of the exit block
        if len(loop.exits) > 1:
            # has multiple exits
            [(returnto, _)] = cfg.successors(
                an_exit
            )  # requirement checked earlier
        else:
            # does not have multiple exits
            returnto = an_exit

        local_block_ids = set(loop.body) | set(loop.entries) | set(loop.exits)
        inputs, outputs = find_region_inout_vars(
            blocks=blocks,
            livemap=livemap,
            callfrom=callfrom,
            returnto=returnto,
            body_block_ids=local_block_ids,
        )

        lli = _loop_lift_info(
            loop=loop,
            inputs=inputs,
            outputs=outputs,
            callfrom=callfrom,
            returnto=returnto,
        )
        loopinfos.append(lli)

    return loopinfos


def _loop_lift_modify_call_block(liftedloop, block, inputs, outputs, returnto):
    """
    Transform calling block from top-level function to call the lifted loop.
    """
    scope = block.scope
    loc = block.loc
    blk = ir.Block(scope=scope, loc=loc)

    ir_utils.fill_block_with_call(
        newblock=blk,
        callee=liftedloop,
        label_next=returnto,
        inputs=inputs,
        outputs=outputs,
    )
    return blk


def _loop_lift_prepare_loop_func(loopinfo, blocks):
    """
    Inplace transform loop blocks for use as lifted loop.
    """
    entry_block = blocks[loopinfo.callfrom]
    scope = entry_block.scope
    loc = entry_block.loc

    # Lowering assumes the first block to be the one with the smallest offset
    firstblk = min(blocks) - 1
    blocks[firstblk] = ir_utils.fill_callee_prologue(
        block=ir.Block(scope=scope, loc=loc),
        inputs=loopinfo.inputs,
        label_next=loopinfo.callfrom,
    )
    blocks[loopinfo.returnto] = ir_utils.fill_callee_epilogue(
        block=ir.Block(scope=scope, loc=loc),
        outputs=loopinfo.outputs,
    )


def _loop_lift_modify_blocks(
    func_ir, loopinfo, blocks, typingctx, targetctx, flags, locals
):
    """
    Modify the block inplace to call to the lifted-loop.
    Returns a dictionary of blocks of the lifted-loop.
    """
    from numba.cuda.dispatcher import LiftedLoop

    # Copy loop blocks
    loop = loopinfo.loop

    loopblockkeys = set(loop.body) | set(loop.entries)
    if len(loop.exits) > 1:
        # has multiple exits
        loopblockkeys |= loop.exits
    loopblocks = dict((k, blocks[k].copy()) for k in loopblockkeys)
    # Modify the loop blocks
    _loop_lift_prepare_loop_func(loopinfo, loopblocks)
    # Since Python 3.13, [END_FOR, POP_TOP] sequence becomes the start of the
    # block causing the block to have line number of the start of previous loop.
    # Fix this using the loc of the first getiter.
    getiter_exprs = []
    for blk in loopblocks.values():
        getiter_exprs.extend(blk.find_exprs(op="getiter"))
    first_getiter = min(getiter_exprs, key=lambda x: x.loc.line)
    loop_loc = first_getiter.loc
    # Create a new IR for the lifted loop
    lifted_ir = func_ir.derive(
        blocks=loopblocks,
        arg_names=tuple(loopinfo.inputs),
        arg_count=len(loopinfo.inputs),
        force_non_generator=True,
        loc=loop_loc,
    )
    liftedloop = LiftedLoop(lifted_ir, typingctx, targetctx, flags, locals)

    # modify for calling into liftedloop
    callblock = _loop_lift_modify_call_block(
        liftedloop,
        blocks[loopinfo.callfrom],
        loopinfo.inputs,
        loopinfo.outputs,
        loopinfo.returnto,
    )
    # remove blocks
    for k in loopblockkeys:
        del blocks[k]
    # update main interpreter callsite into the liftedloop
    blocks[loopinfo.callfrom] = callblock
    return liftedloop


def _has_multiple_loop_exits(cfg, lpinfo):
    """Returns True if there is more than one exit in the loop.

    NOTE: "common exits" refers to the situation where a loop exit has another
    loop exit as its successor. In that case, we do not need to alter it.
    """
    if len(lpinfo.exits) <= 1:
        return False
    exits = set(lpinfo.exits)
    pdom = cfg.post_dominators()

    # Eliminate blocks that have other blocks as post-dominators.
    processed = set()
    remain = set(exits)  # create a copy to work on
    while remain:
        node = remain.pop()
        processed.add(node)
        exits -= pdom[node] - {node}
        remain = exits - processed

    return len(exits) > 1


def _pre_looplift_transform(func_ir):
    """Canonicalize loops for looplifting."""
    from numba.cuda.core.postproc import PostProcessor

    cfg = compute_cfg_from_blocks(func_ir.blocks)
    # For every loop that has multiple exits, combine the exits into one.
    for loop_info in cfg.loops().values():
        if _has_multiple_loop_exits(cfg, loop_info):
            func_ir, _common_key = _fix_multi_exit_blocks(
                func_ir, loop_info.exits
            )
    # Reset and reprocess the func_ir
    func_ir._reset_analysis_variables()
    PostProcessor(func_ir).run()
    return func_ir


def loop_lifting(func_ir, typingctx, targetctx, flags, locals):
    """
    Loop lifting transformation.

    Given a interpreter `func_ir` returns a 2 tuple of
    `(toplevel_interp, [loop0_interp, loop1_interp, ....])`
    """
    func_ir = _pre_looplift_transform(func_ir)
    blocks = func_ir.blocks.copy()
    cfg = compute_cfg_from_blocks(blocks)
    loopinfos = _loop_lift_get_candidate_infos(
        cfg, blocks, func_ir.variable_lifetime.livemap
    )
    loops = []
    if loopinfos:
        _logger.debug(
            "loop lifting this IR with %d candidates:\n%s",
            len(loopinfos),
            func_ir.dump_to_string(),
        )
    for loopinfo in loopinfos:
        lifted = _loop_lift_modify_blocks(
            func_ir, loopinfo, blocks, typingctx, targetctx, flags, locals
        )
        loops.append(lifted)

    # Make main IR
    main = func_ir.derive(blocks=blocks)

    return main, loops


def canonicalize_cfg_single_backedge(blocks):
    """
    Rewrite loops that have multiple backedges.
    """
    cfg = compute_cfg_from_blocks(blocks)
    newblocks = blocks.copy()

    def new_block_id():
        return max(newblocks.keys()) + 1

    def has_multiple_backedges(loop):
        count = 0
        for k in loop.body:
            blk = blocks[k]
            edges = blk.terminator.get_targets()
            # is a backedge?
            if loop.header in edges:
                count += 1
                if count > 1:
                    # early exit
                    return True
        return False

    def yield_loops_with_multiple_backedges():
        for lp in cfg.loops().values():
            if has_multiple_backedges(lp):
                yield lp

    def replace_target(term, src, dst):
        def replace(target):
            return dst if target == src else target

        if isinstance(term, ir.branch_types):
            return ir.Branch(
                cond=term.cond,
                truebr=replace(term.truebr),
                falsebr=replace(term.falsebr),
                loc=term.loc,
            )
        elif isinstance(term, ir.jump_types):
            return ir.Jump(target=replace(term.target), loc=term.loc)
        else:
            assert not term.get_targets()
            return term

    def rewrite_single_backedge(loop):
        """
        Add new tail block that gathers all the backedges
        """
        header = loop.header
        tailkey = new_block_id()
        for blkkey in loop.body:
            blk = newblocks[blkkey]
            if header in blk.terminator.get_targets():
                newblk = blk.copy()
                # rewrite backedge into jumps to new tail block
                newblk.body[-1] = replace_target(
                    blk.terminator, header, tailkey
                )
                newblocks[blkkey] = newblk
        # create new tail block
        entryblk = newblocks[header]
        tailblk = ir.Block(scope=entryblk.scope, loc=entryblk.loc)
        # add backedge
        tailblk.append(ir.Jump(target=header, loc=tailblk.loc))
        newblocks[tailkey] = tailblk

    for loop in yield_loops_with_multiple_backedges():
        rewrite_single_backedge(loop)

    return newblocks


def canonicalize_cfg(blocks):
    """
    Rewrite the given blocks to canonicalize the CFG.
    Returns a new dictionary of blocks.
    """
    return canonicalize_cfg_single_backedge(blocks)


def with_lifting(func_ir, typingctx, targetctx, flags, locals):
    """With-lifting transformation

    Rewrite the IR to extract all withs.
    Only the top-level withs are extracted.
    Returns the (the_new_ir, the_lifted_with_ir)
    """
    from numba.cuda.core import postproc

    def dispatcher_factory(func_ir, objectmode=False, **kwargs):
        from numba.cuda.dispatcher import LiftedWith, ObjModeLiftedWith

        myflags = flags.copy()
        if objectmode:
            # Lifted with-block cannot looplift
            myflags.enable_looplift = False
            # Lifted with-block uses object mode
            myflags.enable_pyobject = True
            myflags.force_pyobject = True
            myflags.no_cpython_wrapper = False
            cls = ObjModeLiftedWith
        else:
            cls = LiftedWith
        return cls(func_ir, typingctx, targetctx, myflags, locals, **kwargs)

    # find where with-contexts regions are
    withs, func_ir = find_setupwiths(func_ir)

    if not withs:
        return func_ir, []

    postproc.PostProcessor(func_ir).run()  # ensure we have variable lifetime
    assert func_ir.variable_lifetime
    vlt = func_ir.variable_lifetime
    blocks = func_ir.blocks.copy()
    cfg = vlt.cfg
    # For each with-regions, mutate them according to
    # the kind of contextmanager
    sub_irs = []
    for blk_start, blk_end in withs:
        body_blocks = _cfg_nodes_in_region(cfg, blk_start, blk_end).copy()
        _legalize_with_head(blocks[blk_start])
        # Find the contextmanager
        cmkind, extra = _get_with_contextmanager(func_ir, blocks, blk_start)
        # Mutate the body and get new IR
        sub = cmkind.mutate_with_body(
            func_ir,
            blocks,
            blk_start,
            blk_end,
            body_blocks,
            dispatcher_factory,
            extra,
        )
        sub_irs.append(sub)
    if not sub_irs:
        # Unchanged
        new_ir = func_ir
    else:
        new_ir = func_ir.derive(blocks)
    return new_ir, sub_irs


def _get_with_contextmanager(func_ir, blocks, blk_start):
    """Get the global object used for the context manager"""
    _illegal_cm_msg = "Illegal use of context-manager."

    def get_var_dfn(var):
        """Get the definition given a variable"""
        return func_ir.get_definition(var)

    def get_ctxmgr_obj(var_ref):
        """Return the context-manager object and extra info.

        The extra contains the arguments if the context-manager is used
        as a call.
        """
        # If the contextmanager used as a Call
        dfn = func_ir.get_definition(var_ref)
        if isinstance(dfn, ir.expr_types) and dfn.op == "call":
            args = [get_var_dfn(x) for x in dfn.args]
            kws = {k: get_var_dfn(v) for k, v in dfn.kws}
            extra = {"args": args, "kwargs": kws}
            var_ref = dfn.func
        else:
            extra = None

        ctxobj = ir_utils.guard(ir_utils.find_outer_value, func_ir, var_ref)

        # check the contextmanager object
        if ctxobj is ir.UNDEFINED:
            raise errors.CompilerError(
                "Undefined variable used as context manager",
                loc=blocks[blk_start].loc,
            )

        if ctxobj is None:
            raise errors.CompilerError(_illegal_cm_msg, loc=dfn.loc)

        return ctxobj, extra

    # Scan the start of the with-region for the contextmanager
    for stmt in blocks[blk_start].body:
        if isinstance(stmt, ir.enterwith_types):
            var_ref = stmt.contextmanager
            ctxobj, extra = get_ctxmgr_obj(var_ref)
            if not hasattr(ctxobj, "mutate_with_body"):
                raise errors.CompilerError(
                    "Unsupported context manager in use",
                    loc=blocks[blk_start].loc,
                )
            return ctxobj, extra
    # No contextmanager found?
    raise errors.CompilerError(
        "malformed with-context usage",
        loc=blocks[blk_start].loc,
    )


def _legalize_with_head(blk):
    """Given *blk*, the head block of the with-context, check that it doesn't
    do anything else.
    """
    counters = defaultdict(int)
    for stmt in blk.body:
        # The counters dict is keyed on the IR node type. As the rest of the
        # function pops out specific node types, we normalize these node types
        # to be the Numba-CUDA versions of them so that we don't have to have
        # more complicated logic looking for both Numba and Numba-CUDA IR nodes
        # of the same kind.
        if isinstance(stmt, ir.enterwith_types):
            stmt_type = ir.EnterWith
        elif isinstance(stmt, ir.jump_types):
            stmt_type = ir.Jump
        elif isinstance(stmt, ir.del_types):
            stmt_type = ir.Del
        else:
            stmt_type = type(stmt)

        counters[stmt_type] += 1

    if counters.pop(ir.EnterWith) != 1:
        raise errors.CompilerError(
            "with's head-block must have exactly 1 ENTER_WITH",
            loc=blk.loc,
        )
    if counters.pop(ir.Jump, 0) != 1:
        raise errors.CompilerError(
            "with's head-block must have exactly 1 JUMP",
            loc=blk.loc,
        )
    # Can have any number of del
    counters.pop(ir.Del, None)
    # There MUST NOT be any other statements
    if counters:
        raise errors.CompilerError(
            "illegal statements in with's head-block",
            loc=blk.loc,
        )


def _cfg_nodes_in_region(cfg, region_begin, region_end):
    """Find the set of CFG nodes that are in the given region"""
    region_nodes = set()
    stack = [region_begin]
    while stack:
        tos = stack.pop()
        succlist = list(cfg.successors(tos))
        # a single block function will have a empty successor list
        if succlist:
            succs, _ = zip(*succlist)
            nodes = set(
                [
                    node
                    for node in succs
                    if node not in region_nodes and node != region_end
                ]
            )
            stack.extend(nodes)
            region_nodes |= nodes

    return region_nodes


def find_setupwiths(func_ir):
    """Find all top-level with.

    Returns a list of ranges for the with-regions.
    """

    def find_ranges(blocks):
        cfg = compute_cfg_from_blocks(blocks)
        sus_setups, sus_pops = set(), set()
        # traverse the cfg and collect all suspected SETUP_WITH and POP_BLOCK
        # statements so that we can iterate over them
        for label, block in blocks.items():
            for stmt in block.body:
                if ir_utils.is_setup_with(stmt):
                    sus_setups.add(label)
                if ir_utils.is_pop_block(stmt):
                    sus_pops.add(label)

        # now that we do have the statements, iterate through them in reverse
        # topo order and from each start looking for pop_blocks
        setup_with_to_pop_blocks_map = defaultdict(set)
        for setup_block in cfg.topo_sort(sus_setups, reverse=True):
            # begin pop_block, search
            to_visit, seen = [], []
            to_visit.append(setup_block)
            while to_visit:
                # get whatever is next and record that we have seen it
                block = to_visit.pop()
                seen.append(block)
                # go through the body of the block, looking for statements
                for stmt in blocks[block].body:
                    # raise detected before pop_block
                    if ir_utils.is_raise(stmt):
                        raise errors.CompilerError(
                            "unsupported control flow due to raise "
                            "statements inside with block"
                        )
                    # if a pop_block, process it
                    if ir_utils.is_pop_block(stmt) and block in sus_pops:
                        # record the jump target of this block belonging to this setup
                        setup_with_to_pop_blocks_map[setup_block].add(block)
                        # remove the block from blocks to be matched
                        sus_pops.remove(block)
                        # stop looking, we have reached the frontier
                        break
                    # if we are still here, by the block terminator,
                    # add all its targets to the to_visit stack, unless we
                    # have seen them already
                    if ir_utils.is_terminator(stmt):
                        to_visit.extend(
                            t for t in stmt.get_targets() if t not in seen
                        )

        return setup_with_to_pop_blocks_map

    blocks = func_ir.blocks
    # initial find, will return a dictionary, mapping indices of blocks
    # containing SETUP_WITH statements to a set of indices of blocks containing
    # POP_BLOCK statements
    with_ranges_dict = find_ranges(blocks)
    # rewrite the CFG in case there are multiple POP_BLOCK statements for one
    # with
    func_ir = consolidate_multi_exit_withs(with_ranges_dict, blocks, func_ir)
    # here we need to turn the withs back into a list of tuples so that the
    # rest of the code can cope
    with_ranges_tuple = [(s, list(p)[0]) for (s, p) in with_ranges_dict.items()]

    # check for POP_BLOCKS with multiple outgoing edges and reject
    for _, p in with_ranges_tuple:
        targets = blocks[p].terminator.get_targets()
        if len(targets) != 1:
            raise errors.CompilerError(
                "unsupported control flow: with-context contains branches "
                "(i.e. break/return/raise) that can leave the block "
            )
    # now we check for returns inside with and reject them
    for _, p in with_ranges_tuple:
        target_block = blocks[p]
        if ir_utils.is_return(
            func_ir.blocks[target_block.terminator.get_targets()[0]].terminator
        ):
            _rewrite_return(func_ir, p)

    # now we need to rewrite the tuple such that we have SETUP_WITH matching the
    # successor of the block that contains the POP_BLOCK.
    with_ranges_tuple = [
        (s, func_ir.blocks[p].terminator.get_targets()[0])
        for (s, p) in with_ranges_tuple
    ]

    # finally we check for nested with statements and reject them
    with_ranges_tuple = _eliminate_nested_withs(with_ranges_tuple)

    return with_ranges_tuple, func_ir


def _rewrite_return(func_ir, target_block_label):
    """Rewrite a return block inside a with statement.

    Arguments
    ---------

    func_ir: Function IR
      the CFG to transform
    target_block_label: int
      the block index/label of the block containing the POP_BLOCK statement


    This implements a CFG transformation to insert a block between two other
    blocks.

    The input situation is:

    ┌───────────────┐
    │   top         │
    │   POP_BLOCK   │
    │   bottom      │
    └───────┬───────┘
            │
    ┌───────▼───────┐
    │               │
    │    RETURN     │
    │               │
    └───────────────┘

    If such a pattern is detected in IR, it means there is a `return` statement
    within a `with` context. The basic idea is to rewrite the CFG as follows:

    ┌───────────────┐
    │   top         │
    │   POP_BLOCK   │
    │               │
    └───────┬───────┘
            │
    ┌───────▼───────┐
    │               │
    │     bottom    │
    │               │
    └───────┬───────┘
            │
    ┌───────▼───────┐
    │               │
    │    RETURN     │
    │               │
    └───────────────┘

    We split the block that contains the `POP_BLOCK` statement into two blocks.
    Everything from the beginning of the block up to and including the
    `POP_BLOCK` statement is considered the 'top' and everything below is
    considered 'bottom'. Finally the jump statements are re-wired to make sure
    the CFG remains valid.

    """
    # the block itself from the index
    target_block = func_ir.blocks[target_block_label]
    # get the index of the block containing the return
    target_block_successor_label = target_block.terminator.get_targets()[0]
    # the return block
    target_block_successor = func_ir.blocks[target_block_successor_label]

    # create the new return block with an appropriate label
    max_label = ir_utils.find_max_label(func_ir.blocks)
    new_label = max_label + 1
    # create the new return block
    new_block_loc = target_block_successor.loc
    new_block_scope = ir.Scope(None, loc=new_block_loc)
    new_block = ir.Block(new_block_scope, loc=new_block_loc)

    # Split the block containing the POP_BLOCK into top and bottom
    # Block must be of the form:
    # -----------------
    # <some stmts>
    # POP_BLOCK
    # <some more stmts>
    # JUMP
    # -----------------
    top_body, bottom_body = [], []
    pop_blocks = [*target_block.find_insts(ir.popblock_types)]
    assert len(pop_blocks) == 1
    assert len([*target_block.find_insts(ir.jump_types)]) == 1
    assert isinstance(target_block.body[-1], ir.jump_types)
    pb_marker = pop_blocks[0]
    pb_is = target_block.body.index(pb_marker)
    top_body.extend(target_block.body[:pb_is])
    top_body.append(ir.Jump(target_block_successor_label, target_block.loc))
    bottom_body.extend(target_block.body[pb_is:-1])
    bottom_body.append(ir.Jump(new_label, target_block.loc))

    # get the contents of the return block
    return_body = func_ir.blocks[target_block_successor_label].body
    # finally, re-assign all blocks
    new_block.body.extend(return_body)
    target_block_successor.body.clear()
    target_block_successor.body.extend(bottom_body)
    target_block.body.clear()
    target_block.body.extend(top_body)

    # finally, append the new return block and rebuild the IR properties
    func_ir.blocks[new_label] = new_block
    func_ir._definitions = ir_utils.build_definitions(func_ir.blocks)
    return func_ir


def _eliminate_nested_withs(with_ranges):
    known_ranges = []

    def within_known_range(start, end, known_ranges):
        for a, b in known_ranges:
            # FIXME: this should be a comparison in topological order, right
            # now we are comparing the integers of the blocks, stuff probably
            # works by accident.
            if start > a and end < b:
                return True
        return False

    for s, e in sorted(with_ranges):
        if not within_known_range(s, e, known_ranges):
            known_ranges.append((s, e))

    return known_ranges


def consolidate_multi_exit_withs(withs: dict, blocks, func_ir):
    """Modify the FunctionIR to merge the exit blocks of with constructs."""
    for k in withs:
        vs: set = withs[k]
        if len(vs) > 1:
            func_ir, common = _fix_multi_exit_blocks(
                func_ir,
                vs,
                split_condition=ir_utils.is_pop_block,
            )
            withs[k] = {common}
    return func_ir


def _fix_multi_exit_blocks(func_ir, exit_nodes, *, split_condition=None):
    """Modify the FunctionIR to create a single common exit node given the
    original exit nodes.

    Parameters
    ----------
    func_ir :
        The FunctionIR. Mutated inplace.
    exit_nodes :
        The original exit nodes. A sequence of block keys.
    split_condition : callable or None
        If not None, it is a callable with the signature
        `split_condition(statement)` that determines if the `statement` is the
        splitting point (e.g. `POP_BLOCK`) in an exit node.
        If it's None, the exit node is not split.
    """

    # Convert the following:
    #
    #     |           |
    # +-------+   +-------+
    # | exit0 |   | exit1 |
    # +-------+   +-------+
    #     |           |
    # +-------+   +-------+
    # | after0|   | after1|
    # +-------+   +-------+
    #     |           |
    #
    # To roughly:
    #
    #     |           |
    # +-------+   +-------+
    # | exit0 |   | exit1 |
    # +-------+   +-------+
    #     |           |
    #     +-----+-----+
    #           |
    #      +---------+
    #      | common  |
    #      +---------+
    #           |
    #       +-------+
    #       | post  |
    #       +-------+
    #           |
    #     +-----+-----+
    #     |           |
    # +-------+   +-------+
    # | after0|   | after1|
    # +-------+   +-------+

    blocks = func_ir.blocks
    # Getting the scope
    any_blk = min(func_ir.blocks.values())
    scope = any_blk.scope
    # Getting the maximum block label
    max_label = max(func_ir.blocks) + 1
    # Define the new common block for the new exit.
    common_block = ir.Block(any_blk.scope, loc=ir.unknown_loc)
    common_label = max_label
    max_label += 1
    blocks[common_label] = common_block
    # Define the new block after the exit.
    post_block = ir.Block(any_blk.scope, loc=ir.unknown_loc)
    post_label = max_label
    max_label += 1
    blocks[post_label] = post_block

    # Adjust each exit node
    remainings = []
    for i, k in enumerate(exit_nodes):
        blk = blocks[k]

        # split the block if needed
        if split_condition is not None:
            for stmt in blk.body:
                if split_condition(stmt):
                    break
        else:
            # no splitting
            pt = -1

        before = blk.body[:pt]
        after = blk.body[pt:]
        remainings.append(after)

        # Add control-point variable to mark which exit block this is.
        blk.body = before
        loc = blk.loc
        blk.body.append(
            ir.Assign(
                value=ir.Const(i, loc=loc),
                target=scope.get_or_define("$cp", loc=loc),
                loc=loc,
            )
        )
        # Replace terminator with a jump to the common block
        assert not blk.is_terminated
        blk.body.append(ir.Jump(common_label, loc=ir.unknown_loc))

    if split_condition is not None:
        # Move the splitting statement to the common block
        common_block.body.append(remainings[0][0])
    assert not common_block.is_terminated
    # Append jump from common block to post block
    common_block.body.append(ir.Jump(post_label, loc=loc))

    # Make if-else tree to jump to target
    remain_blocks = range(max_label, max_label + len(remainings))

    switch_block = post_block
    loc = ir.unknown_loc
    for i, remain in enumerate(remainings):
        match_expr = scope.redefine("$cp_check", loc=loc)
        match_rhs = scope.redefine("$cp_rhs", loc=loc)

        # Do comparison to match control-point variable to the exit block
        switch_block.body.append(
            ir.Assign(value=ir.Const(i, loc=loc), target=match_rhs, loc=loc),
        )

        # Add assignment for the comparison
        switch_block.body.append(
            ir.Assign(
                value=ir.Expr.binop(
                    fn=operator.eq,
                    lhs=scope.get("$cp"),
                    rhs=match_rhs,
                    loc=loc,
                ),
                target=match_expr,
                loc=loc,
            ),
        )

        # Insert jump to the next case
        [jump_target] = remain[-1].get_targets()
        switch_block.body.append(
            ir.Branch(match_expr, jump_target, remain_blocks[i], loc=loc),
        )
        switch_block = ir.Block(scope=scope, loc=loc)
        blocks[remain_blocks[i]] = switch_block

    # Add the final jump
    switch_block.body.append(ir.Jump(jump_target, loc=loc))

    return func_ir, common_label
